# -*- coding: utf-8 -*-
import random
import string

# Py3 compatibility
import sys
import unittest

from nectar import Hive
from nectar.account import Account
from nectar.amount import Amount
from nectar.exceptions import InvalidWifError, MissingKeyError
from nectar.instance import shared_blockchain_instance
from nectar.memo import Memo
from nectar.nodelist import NodeList
from nectar.transactionbuilder import TransactionBuilder
from nectarapi import exceptions
from nectarbase.operations import Transfer
from nectargraphenebase.account import PrivateKey, PublicKey

core_unit = "STX"


class Testcases(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        nodelist = NodeList()
        # hv = shared_blockchain_instance()
        # hv.config.refreshBackup()
        # nodes = nodelist.get_testnet()
        cls.nodes = nodelist.get_nodes()
        cls.bts = Hive(
            node=cls.nodes,
            nobroadcast=True,
            num_retries=10,
            expiration=120,
        )
        # from getpass import getpass
        # self.bts.wallet.unlock(getpass())
        cls.bts.set_default_account("nectar")

        # Test account "nectar"
        cls.active_key = "5Jt2wTfhUt5GkZHV1HYVfkEaJ6XnY8D2iA4qjtK9nnGXAhThM3w"
        cls.posting_key = "5Jh1Gtu2j4Yi16TfhoDmg8Qj3ULcgRi7A49JXdfUUTVPkaFaRKz"
        cls.memo_key = "5KPbCuocX26aMxN9CDPdUex4wCbfw9NoT5P7UhcqgDwxXa47bit"

        # Test account "nectar1"
        cls.active_key1 = "5Jo9SinzpdAiCDLDJVwuN7K5JcusKmzFnHpEAtPoBHaC1B5RDUd"
        cls.posting_key1 = "5JGNhDXuDLusTR3nbmpWAw4dcmE8WfSM8odzqcQ6mDhJHP8YkQo"
        cls.memo_key1 = "5KA2ddfAffjfRFoe1UhQjJtKnGsBn9xcsdPQTfMt1fQuErDAkWr"

        cls.active_private_key_of_nectar4 = "5JkZZEUWrDsu3pYF7aknSo7BLJx7VfxB3SaRtQaHhsPouDYjxzi"
        cls.active_private_key_of_nectar5 = "5Hvbm9VjRbd1B3ft8Lm81csaqQudwFwPGdiRKrCmTKcomFS3Z9J"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        raise unittest.SkipTest()
        hv = self.bts
        hv.nobroadcast = True
        hv.wallet.wipe(True)
        hv.wallet.create("123")
        hv.wallet.unlock("123")

        hv.wallet.addPrivateKey(self.active_key1)
        hv.wallet.addPrivateKey(self.memo_key1)
        hv.wallet.addPrivateKey(self.posting_key1)

        hv.wallet.addPrivateKey(self.active_key)
        hv.wallet.addPrivateKey(self.memo_key)
        hv.wallet.addPrivateKey(self.posting_key)
        hv.wallet.addPrivateKey(self.active_private_key_of_nectar4)
        hv.wallet.addPrivateKey(self.active_private_key_of_nectar5)

    @classmethod
    def tearDownClass(cls):
        hv = shared_blockchain_instance()
        hv.config.recover_with_latest_backup()

    def test_wallet_keys(self):
        hv = self.bts
        hv.wallet.unlock("123")
        priv_key = hv.wallet.getPrivateKeyForPublicKey(
            str(PrivateKey(self.posting_key, prefix=hv.prefix).pubkey)
        )
        self.assertEqual(str(priv_key), self.posting_key)
        priv_key = hv.wallet.getKeyForAccount("nectar", "active")
        self.assertEqual(str(priv_key), self.active_key)
        priv_key = hv.wallet.getKeyForAccount("nectar1", "posting")
        self.assertEqual(str(priv_key), self.posting_key1)

        priv_key = hv.wallet.getPrivateKeyForPublicKey(
            str(PrivateKey(self.active_private_key_of_nectar4, prefix=hv.prefix).pubkey)
        )
        self.assertEqual(str(priv_key), self.active_private_key_of_nectar4)
        priv_key = hv.wallet.getKeyForAccount("nectar4", "active")
        self.assertEqual(str(priv_key), self.active_private_key_of_nectar4)

        priv_key = hv.wallet.getPrivateKeyForPublicKey(
            str(PrivateKey(self.active_private_key_of_nectar5, prefix=hv.prefix).pubkey)
        )
        self.assertEqual(str(priv_key), self.active_private_key_of_nectar5)
        priv_key = hv.wallet.getKeyForAccount("nectar5", "active")
        self.assertEqual(str(priv_key), self.active_private_key_of_nectar5)

    def test_transfer(self):
        bts = self.bts
        bts.nobroadcast = False
        bts.wallet.unlock("123")
        # bts.wallet.addPrivateKey(self.active_key)
        # bts.prefix ="STX"
        acc = Account("nectar", blockchain_instance=bts)
        tx = acc.transfer("nectar1", 1.33, "HBD", memo="Foobar")
        self.assertEqual(tx["operations"][0][0], "transfer")
        self.assertEqual(len(tx["signatures"]), 1)
        op = tx["operations"][0][1]
        self.assertIn("memo", op)
        self.assertEqual(op["from"], "nectar")
        self.assertEqual(op["to"], "nectar1")
        amount = Amount(op["amount"], blockchain_instance=bts)
        self.assertEqual(float(amount), 1.33)
        bts.nobroadcast = True

    def test_transfer_memo(self):
        bts = self.bts
        bts.nobroadcast = False
        bts.wallet.unlock("123")
        acc = Account("nectar", blockchain_instance=bts)
        tx = acc.transfer("nectar1", 1.33, "HBD", memo="#Foobar")
        self.assertEqual(tx["operations"][0][0], "transfer")
        op = tx["operations"][0][1]
        self.assertIn("memo", op)
        self.assertIn("#", op["memo"])
        m = Memo(from_account=op["from"], to_account=op["to"], blockchain_instance=bts)
        memo = m.decrypt(op["memo"])
        self.assertEqual(memo, "Foobar")

        self.assertEqual(op["from"], "nectar")
        self.assertEqual(op["to"], "nectar1")
        amount = Amount(op["amount"], blockchain_instance=bts)
        self.assertEqual(float(amount), 1.33)
        bts.nobroadcast = True

    def test_transfer_1of1(self):
        hive = self.bts
        hive.nobroadcast = False
        tx = TransactionBuilder(use_condenser_api=True, blockchain_instance=hive)
        tx.appendOps(
            Transfer(
                **{
                    "from": "nectar",
                    "to": "nectar1",
                    "amount": Amount("0.01 STEEM", blockchain_instance=hive),
                    "memo": "1 of 1 transaction",
                }
            )
        )
        self.assertEqual(tx["operations"][0]["type"], "transfer_operation")
        tx.appendWif(self.active_key)
        tx.sign()
        tx.sign()
        self.assertEqual(len(tx["signatures"]), 1)
        tx.broadcast()
        hive.nobroadcast = True

    def test_transfer_2of2_simple(self):
        # Send a 2 of 2 transaction from elf which needs nectar4's cosign to send funds
        hive = self.bts
        hive.nobroadcast = False
        tx = TransactionBuilder(use_condenser_api=True, blockchain_instance=hive)
        tx.appendOps(
            Transfer(
                **{
                    "from": "nectar5",
                    "to": "nectar1",
                    "amount": Amount("0.01 HIVE", blockchain_instance=hive),
                    "memo": "2 of 2 simple transaction",
                }
            )
        )

        tx.appendWif(self.active_private_key_of_nectar5)
        tx.sign()
        tx.clearWifs()
        tx.appendWif(self.active_private_key_of_nectar4)
        tx.sign(reconstruct_tx=False)
        self.assertEqual(len(tx["signatures"]), 2)
        tx.broadcast()
        hive.nobroadcast = True

    def test_transfer_2of2_wallet(self):
        # Send a 2 of 2 transaction from nectar5 which needs nectar4's cosign to send
        # priv key of nectar5 and nectar4 are stored in the wallet
        # appendSigner fetches both keys and signs automatically with both keys.
        hive = self.bts
        hive.nobroadcast = False
        hive.wallet.unlock("123")

        tx = TransactionBuilder(use_condenser_api=True, blockchain_instance=hive)
        tx.appendOps(
            Transfer(
                **{
                    "from": "nectar5",
                    "to": "nectar1",
                    "amount": Amount("0.01 HIVE", blockchain_instance=hive),
                    "memo": "2 of 2 serialized/deserialized transaction",
                }
            )
        )

        tx.appendSigner("nectar5", "active")
        tx.sign()
        self.assertEqual(len(tx["signatures"]), 2)
        tx.broadcast()
        hive.nobroadcast = True

    def test_transfer_2of2_serialized_deserialized(self):
        # Send a 2 of 2 transaction from nectar5 which needs nectar4's cosign to send
        # funds but sign the transaction with nectar5's key and then serialize the transaction
        # and deserialize the transaction.  After that, sign with nectar4's key.
        hive = self.bts
        hive.nobroadcast = False
        hive.wallet.unlock("123")
        # hive.wallet.removeAccount("nectar4")
        hive.wallet.removePrivateKeyFromPublicKey(
            str(PublicKey(self.active_private_key_of_nectar4, prefix=core_unit))
        )

        tx = TransactionBuilder(use_condenser_api=True, blockchain_instance=hive)
        tx.appendOps(
            Transfer(
                **{
                    "from": "nectar5",
                    "to": "nectar1",
                    "amount": Amount("0.01 STEEM", blockchain_instance=hive),
                    "memo": "2 of 2 serialized/deserialized transaction",
                }
            )
        )

        tx.appendSigner("nectar5", "active")
        tx.addSigningInformation("nectar5", "active")
        tx.sign()
        tx.clearWifs()
        self.assertEqual(len(tx["signatures"]), 1)
        # hive.wallet.removeAccount("nectar5")
        hive.wallet.removePrivateKeyFromPublicKey(
            str(PublicKey(self.active_private_key_of_nectar5, prefix=core_unit))
        )
        tx_json = tx.json()
        del tx
        new_tx = TransactionBuilder(tx=tx_json, blockchain_instance=hive)
        self.assertEqual(len(new_tx["signatures"]), 1)
        hive.wallet.addPrivateKey(self.active_private_key_of_nectar4)
        new_tx.appendMissingSignatures()
        new_tx.sign(reconstruct_tx=False)
        self.assertEqual(len(new_tx["signatures"]), 2)
        new_tx.broadcast()
        hive.nobroadcast = True

    def test_transfer_2of2_offline(self):
        # Send a 2 of 2 transaction from nectar5 which needs nectar4's cosign to send
        # funds but sign the transaction with nectar5's key and then serialize the transaction
        # and deserialize the transaction.  After that, sign with nectar4's key.
        hive = self.bts
        hive.nobroadcast = False
        hive.wallet.unlock("123")
        # hive.wallet.removeAccount("nectar4")
        hive.wallet.removePrivateKeyFromPublicKey(
            str(PublicKey(self.active_private_key_of_nectar4, prefix=core_unit))
        )

        tx = TransactionBuilder(use_condenser_api=True, blockchain_instance=hive)
        tx.appendOps(
            Transfer(
                **{
                    "from": "nectar5",
                    "to": "nectar",
                    "amount": Amount("0.01 STEEM", blockchain_instance=hive),
                    "memo": "2 of 2 serialized/deserialized transaction",
                }
            )
        )

        tx.appendSigner("nectar5", "active")
        tx.addSigningInformation("nectar5", "active")
        tx.sign()
        tx.clearWifs()
        self.assertEqual(len(tx["signatures"]), 1)
        # hive.wallet.removeAccount("nectar5")
        hive.wallet.removePrivateKeyFromPublicKey(
            str(PublicKey(self.active_private_key_of_nectar5, prefix=core_unit))
        )
        hive.wallet.addPrivateKey(self.active_private_key_of_nectar4)
        tx.appendMissingSignatures()
        tx.sign(reconstruct_tx=False)
        self.assertEqual(len(tx["signatures"]), 2)
        tx.broadcast()
        hive.nobroadcast = True
        hive.wallet.addPrivateKey(self.active_private_key_of_nectar5)

    def test_transfer_2of2_wif(self):
        _ = NodeList()
        # Send a 2 of 2 transaction from elf which needs nectar4's cosign to send
        # funds but sign the transaction with elf's key and then serialize the transaction
        # and deserialize the transaction.  After that, sign with nectar4's key.
        hive = Hive(
            node=self.nodes,
            num_retries=10,
            keys=[self.active_private_key_of_nectar5],
            expiration=360,
        )

        tx = TransactionBuilder(use_condenser_api=True, blockchain_instance=hive)
        tx.appendOps(
            Transfer(
                **{
                    "from": "nectar5",
                    "to": "nectar",
                    "amount": Amount("0.01 HIVE", blockchain_instance=hive),
                    "memo": "2 of 2 serialized/deserialized transaction",
                }
            )
        )

        tx.appendSigner("nectar5", "active")
        tx.addSigningInformation("nectar5", "active")
        tx.sign()
        tx.clearWifs()
        self.assertEqual(len(tx["signatures"]), 1)
        tx_json = tx.json()
        del hive
        del tx

        hive = Hive(
            node=self.nodes,
            num_retries=10,
            keys=[self.active_private_key_of_nectar4],
            expiration=360,
        )
        new_tx = TransactionBuilder(tx=tx_json, blockchain_instance=hive)
        new_tx.appendMissingSignatures()
        new_tx.sign(reconstruct_tx=False)
        self.assertEqual(len(new_tx["signatures"]), 2)
        new_tx.broadcast()

    def test_verifyAuthority(self):
        hv = self.bts
        hv.wallet.unlock("123")
        tx = TransactionBuilder(use_condenser_api=True, blockchain_instance=hv)
        tx.appendOps(
            Transfer(
                **{
                    "from": "nectar",
                    "to": "nectar1",
                    "amount": Amount("1.300 HBD", blockchain_instance=hv),
                    "memo": "Foobar",
                }
            )
        )
        account = Account("nectar", blockchain_instance=hv)
        tx.appendSigner(account, "active")
        self.assertTrue(len(tx.wifs) > 0)
        tx.sign()
        tx.verify_authority()
        self.assertTrue(len(tx["signatures"]) > 0)

    def test_create_account(self):
        bts = self.bts
        name = "".join(random.choice(string.ascii_lowercase) for _ in range(12))
        key1 = PrivateKey()
        key2 = PrivateKey()
        key3 = PrivateKey()
        key4 = PrivateKey()
        key5 = PrivateKey()
        tx = bts.create_account(
            name,
            creator="nectar",
            owner_key=format(key1.pubkey, core_unit),
            active_key=format(key2.pubkey, core_unit),
            posting_key=format(key3.pubkey, core_unit),
            memo_key=format(key4.pubkey, core_unit),
            additional_owner_keys=[format(key5.pubkey, core_unit)],
            additional_active_keys=[format(key5.pubkey, core_unit)],
            additional_owner_accounts=["nectar1"],  # 1.2.0
            additional_active_accounts=["nectar1"],
            storekeys=False,
        )
        self.assertEqual(tx["operations"][0][0], "account_create")
        op = tx["operations"][0][1]
        role = "active"
        self.assertIn(format(key5.pubkey, core_unit), [x[0] for x in op[role]["key_auths"]])
        self.assertIn(format(key5.pubkey, core_unit), [x[0] for x in op[role]["key_auths"]])
        self.assertIn("nectar1", [x[0] for x in op[role]["account_auths"]])
        role = "owner"
        self.assertIn(format(key5.pubkey, core_unit), [x[0] for x in op[role]["key_auths"]])
        self.assertIn(format(key5.pubkey, core_unit), [x[0] for x in op[role]["key_auths"]])
        self.assertIn("nectar1", [x[0] for x in op[role]["account_auths"]])
        self.assertEqual(op["creator"], "nectar")

    def test_connect(self):
        _ = NodeList()
        self.bts.connect(node=self.nodes)
        bts = self.bts
        self.assertEqual(bts.prefix, "STX")

    def test_set_default_account(self):
        self.bts.set_default_account("nectar")

    def test_info(self):
        info = self.bts.info()
        for key in [
            "current_witness",
            "head_block_id",
            "head_block_number",
            "id",
            "last_irreversible_block_num",
            "current_witness",
            "total_pow",
            "time",
        ]:
            self.assertTrue(key in info)

    def test_finalizeOps(self):
        bts = self.bts
        tx1 = bts.new_tx()
        tx2 = bts.new_tx()

        acc = Account("nectar", blockchain_instance=bts)
        acc.transfer("nectar1", 1, "STEEM", append_to=tx1)
        acc.transfer("nectar1", 2, "STEEM", append_to=tx2)
        acc.transfer("nectar1", 3, "STEEM", append_to=tx1)
        tx1 = tx1.json()
        tx2 = tx2.json()
        ops1 = tx1["operations"]
        ops2 = tx2["operations"]
        self.assertEqual(len(ops1), 2)
        self.assertEqual(len(ops2), 1)

    def test_weight_threshold(self):
        bts = self.bts
        auth = {
            "account_auths": [["test", 1]],
            "extensions": [],
            "key_auths": [
                ["STX55VCzsb47NZwWe5F3qyQKedX9iHBHMVVFSc96PDvV7wuj7W86n", 1],
                ["STX7GM9YXcsoAJAgKbqW2oVj7bnNXFNL4pk9NugqKWPmuhoEDbkDv", 1],
            ],
            "weight_threshold": 3,
        }  # threshold fine
        bts._test_weights_treshold(auth)
        auth = {
            "account_auths": [["test", 1]],
            "extensions": [],
            "key_auths": [
                ["STX55VCzsb47NZwWe5F3qyQKedX9iHBHMVVFSc96PDvV7wuj7W86n", 1],
                ["STX7GM9YXcsoAJAgKbqW2oVj7bnNXFNL4pk9NugqKWPmuhoEDbkDv", 1],
            ],
            "weight_threshold": 4,
        }  # too high

        with self.assertRaises(ValueError):
            bts._test_weights_treshold(auth)

    def test_allow(self):
        bts = self.bts
        self.assertIn(bts.prefix, "STX")
        acc = Account("nectar", blockchain_instance=bts)
        self.assertIn(acc.hive.prefix, "STX")
        tx = acc.allow(
            "STX55VCzsb47NZwWe5F3qyQKedX9iHBHMVVFSc96PDvV7wuj7W86n",
            account="nectar",
            weight=1,
            threshold=1,
            permission="active",
        )
        self.assertEqual((tx["operations"][0][0]), "account_update")
        op = tx["operations"][0][1]
        self.assertIn("active", op)
        self.assertIn(
            ["STX55VCzsb47NZwWe5F3qyQKedX9iHBHMVVFSc96PDvV7wuj7W86n", "1"],
            op["active"]["key_auths"],
        )
        self.assertEqual(op["active"]["weight_threshold"], 1)

    def test_disallow(self):
        bts = self.bts
        acc = Account("nectar", blockchain_instance=bts)
        if sys.version > "3":
            _assertRaisesRegex = self.assertRaisesRegex
        else:
            _assertRaisesRegex = self.assertRaisesRegexp
        with _assertRaisesRegex(ValueError, ".*Changes nothing.*"):
            acc.disallow(
                "STX55VCzsb47NZwWe5F3qyQKedX9iHBHMVVFSc96PDvV7wuj7W86n",
                weight=1,
                threshold=1,
                permission="active",
            )
        with _assertRaisesRegex(ValueError, ".*Changes nothing!.*"):
            acc.disallow(
                "STX6MRyAjQq8ud7hVNYcfnVPJqcVpscN5So8BhtHuGYqET5GDW5CV",
                weight=1,
                threshold=1,
                permission="active",
            )

    def test_update_memo_key(self):
        bts = self.bts
        bts.wallet.unlock("123")
        self.assertEqual(bts.prefix, "STX")
        acc = Account("nectar", blockchain_instance=bts)
        tx = acc.update_memo_key("STX55VCzsb47NZwWe5F3qyQKedX9iHBHMVVFSc96PDvV7wuj7W86n")
        self.assertEqual((tx["operations"][0][0]), "account_update")
        op = tx["operations"][0][1]
        self.assertEqual(op["memo_key"], "STX55VCzsb47NZwWe5F3qyQKedX9iHBHMVVFSc96PDvV7wuj7W86n")

    def test_approvewitness(self):
        bts = self.bts
        w = Account("nectar", blockchain_instance=bts)
        tx = w.approvewitness("nectar1")
        self.assertEqual((tx["operations"][0][0]), "account_witness_vote")
        op = tx["operations"][0][1]
        self.assertIn("nectar1", op["witness"])

    def test_appendWif(self):
        _ = NodeList()
        hv = Hive(node=self.nodes, nobroadcast=True, expiration=120, num_retries=10)
        tx = TransactionBuilder(use_condenser_api=True, blockchain_instance=hv)
        tx.appendOps(
            Transfer(
                **{
                    "from": "nectar",
                    "to": "nectar1",
                    "amount": Amount("1 STEEM", blockchain_instance=hv),
                    "memo": "",
                }
            )
        )
        with self.assertRaises(MissingKeyError):
            tx.sign()
        with self.assertRaises(InvalidWifError):
            tx.appendWif("abcdefg")
        tx.appendWif(self.active_key)
        tx.sign()
        self.assertTrue(len(tx["signatures"]) > 0)

    def test_appendSigner(self):
        _ = NodeList()
        hv = Hive(
            node=self.nodes,
            keys=[self.active_key],
            nobroadcast=True,
            expiration=120,
            num_retries=10,
        )
        tx = TransactionBuilder(use_condenser_api=True, blockchain_instance=hv)
        tx.appendOps(
            Transfer(
                **{
                    "from": "nectar",
                    "to": "nectar1",
                    "amount": Amount("1 STEEM", blockchain_instance=hv),
                    "memo": "",
                }
            )
        )
        account = Account("nectar", blockchain_instance=hv)
        with self.assertRaises(AssertionError):
            tx.appendSigner(account, "abcdefg")
        tx.appendSigner(account, "active")
        self.assertTrue(len(tx.wifs) > 0)
        tx.sign()
        self.assertTrue(len(tx["signatures"]) > 0)

    def test_verifyAuthorityException(self):
        _ = NodeList()
        hv = Hive(
            node=self.nodes,
            keys=[self.posting_key],
            nobroadcast=True,
            expiration=120,
            num_retries=10,
        )
        tx = TransactionBuilder(use_condenser_api=True, blockchain_instance=hv)
        tx.appendOps(
            Transfer(
                **{
                    "from": "nectar",
                    "to": "nectar1",
                    "amount": Amount("1 STEEM", blockchain_instance=hv),
                    "memo": "",
                }
            )
        )
        account = Account("nectar2", blockchain_instance=hv)
        tx.appendSigner(account, "active")
        tx.appendWif(self.posting_key)
        self.assertTrue(len(tx.wifs) > 0)
        tx.sign()
        with self.assertRaises(exceptions.MissingRequiredActiveAuthority):
            tx.verify_authority()
        self.assertTrue(len(tx["signatures"]) > 0)

    def test_Transfer_broadcast(self):
        _ = NodeList()
        hv = Hive(
            node=self.nodes,
            keys=[self.active_key],
            nobroadcast=True,
            expiration=120,
            num_retries=10,
        )

        tx = TransactionBuilder(use_condenser_api=True, expiration=10, blockchain_instance=hv)
        tx.appendOps(
            Transfer(
                **{
                    "from": "nectar",
                    "to": "nectar1",
                    "amount": Amount("1 STEEM", blockchain_instance=hv),
                    "memo": "",
                }
            )
        )
        tx.appendSigner("nectar", "active")
        tx.sign()
        tx.broadcast()

    def test_TransactionConstructor(self):
        hv = self.bts
        opTransfer = Transfer(
            **{
                "from": "nectar",
                "to": "nectar1",
                "amount": Amount("1 STEEM", blockchain_instance=hv),
                "memo": "",
            }
        )
        tx1 = TransactionBuilder(use_condenser_api=True, blockchain_instance=hv)
        tx1.appendOps(opTransfer)
        tx = TransactionBuilder(tx1, blockchain_instance=hv)
        self.assertFalse(tx.is_empty())
        self.assertTrue(len(tx.list_operations()) == 1)
        self.assertTrue(repr(tx) is not None)
        self.assertTrue(str(tx) is not None)
        account = Account("nectar", blockchain_instance=hv)
        tx.appendSigner(account, "active")
        self.assertTrue(len(tx.wifs) > 0)
        tx.sign()
        self.assertTrue(len(tx["signatures"]) > 0)

    def test_follow_active_key(self):
        _ = NodeList()
        hv = Hive(
            node=self.nodes,
            keys=[self.active_key],
            nobroadcast=True,
            expiration=120,
            num_retries=10,
        )
        account = Account("nectar", blockchain_instance=hv)
        account.follow("nectar1")

    def test_follow_posting_key(self):
        _ = NodeList()
        hv = Hive(
            node=self.nodes,
            keys=[self.posting_key],
            nobroadcast=True,
            expiration=120,
            num_retries=10,
        )
        account = Account("nectar", blockchain_instance=hv)
        account.follow("nectar1")



from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import time
from datetime import datetime, timedelta

from nectar.block import Block
from nectar.blockchain import Blockchain
from nectar.hive import Hive
from nectar.nodelist import NodeList
from nectar.utils import formatTimedelta

log = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


if __name__ == "__main__":
    node_setup = 1
    how_many_hours = 1
    nodes = NodeList()
    if node_setup == 0:
        hv = Hive(node=nodes.get_nodes(normal=True, https=True), num_retries=10)
        max_batch_size = None
        threading = False
        thread_num = 8
    elif node_setup == 1:
        hv = Hive(node=nodes.get_nodes(normal=True, https=True), num_retries=10)
        max_batch_size = None
        threading = True
        thread_num = 16
    elif node_setup == 2:
        hv = Hive(node=nodes.get_nodes(appbase=False, https=True), num_retries=10)
        max_batch_size = None
        threading = True
        thread_num = 16
    blockchain = Blockchain(blockchain_instance=hv)
    last_block_id = 19273700
    last_block = Block(last_block_id, blockchain_instance=hv)
    startTime = datetime.now()

    stopTime = last_block.time() + timedelta(seconds=how_many_hours * 60 * 60)
    ltime = time.time()
    cnt = 0
    total_transaction = 0

    start_time = time.time()
    last_node = hv.rpc.url
    print("Current node:", last_node)
    for entry in blockchain.blocks(
        start=last_block_id,
        max_batch_size=max_batch_size,
        threading=threading,
        thread_num=thread_num,
        thread_limit=1200,
    ):
        block_no = entry.identifier
        if "block" in entry:
            trxs = entry["block"]["transactions"]
        else:
            trxs = entry["transactions"]

        for tx in trxs:
            for op in tx["operations"]:
                total_transaction += 1
        if "block" in entry:
            block_time = entry["block"]["timestamp"]
        else:
            block_time = entry["timestamp"]

        if block_time > stopTime:
            total_duration = formatTimedelta(datetime.now() - startTime)
            last_block_id = block_no
            avtran = total_transaction / (last_block_id - 19273700)
            print(
                "* HOUR mark: Processed %d blockchain hours in %s"
                % (how_many_hours, total_duration)
            )
            print(
                "* Blocks %d, Transactions %d (Avg. per Block %f)"
                % ((last_block_id - 19273700), total_transaction, avtran)
            )
            break

        if block_no != last_block_id:
            cnt += 1
            last_block_id = block_no
            if last_block_id % 100 == 0:
                now = time.time()
                duration = now - ltime
                total_duration = now - start_time
                speed = int(100000.0 / duration) * 1.0 / 1000
                avspeed = int((last_block_id - 19273700) * 1000 / total_duration) * 1.0 / 1000
                avtran = total_transaction / (last_block_id - 19273700)
                ltime = now
                if last_node != hv.rpc.url:
                    last_node = hv.rpc.url
                    print("Current node:", last_node)
                print(
                    "* 100 blocks processed in %.2f seconds. Speed %.2f. Avg: %.2f. Avg.Trans:"
                    "%.2f Count: %d Block minutes: %d"
                    % (duration, speed, avspeed, avtran, cnt, cnt * 3 / 60)
                )



# -*- coding: utf-8 -*-
from decimal import Decimal
from fractions import Fraction

from nectar.instance import shared_blockchain_instance

from .amount import Amount, check_asset
from .asset import Asset
from .exceptions import InvalidAssetException
from .utils import assets_from_string, formatTimeString


class Price(dict):
    """This class deals with all sorts of prices of any pair of assets to
    simplify dealing with the tuple::

        (quote, base)

    each being an instance of :class:`nectar.amount.Amount`. The
    amount themselves define the price.

    .. note::

        The price (floating) is derived as ``base/quote``

    :param list args: Allows to deal with different representations of a price
    :param Asset base: Base asset
    :param Asset quote: Quote asset
    :param Hive blockchain_instance: Hive instance
    :returns: All data required to represent a price
    :rtype: dictionary

    Way to obtain a proper instance:

        * ``args`` is a str with a price and two assets
        * ``args`` can be a floating number and ``base`` and ``quote`` being instances of :class:`nectar.asset.Asset`
        * ``args`` can be a floating number and ``base`` and ``quote`` being instances of ``str``
        * ``args`` can be dict with keys ``price``, ``base``, and ``quote`` (*graphene balances*)
        * ``args`` can be dict with keys ``base`` and ``quote``
        * ``args`` can be dict with key ``receives`` (filled orders)
        * ``args`` being a list of ``[quote, base]`` both being instances of :class:`nectar.amount.Amount`
        * ``args`` being a list of ``[quote, base]`` both being instances of ``str`` (``amount symbol``)
        * ``base`` and ``quote`` being instances of :class:`nectar.asset.Amount`

    This allows instantiations like:

    * ``Price("0.315 HBD/HIVE")``
    * ``Price(0.315, base="HBD", quote="HIVE")``
    * ``Price(0.315, base=Asset("HBD"), quote=Asset("HIVE"))``
    * ``Price({"base": {"amount": 1, "asset_id": "HBD"}, "quote": {"amount": 10, "asset_id": "HBD"}})``
    * ``Price(quote="10 HIVE", base="1 HBD")``
    * ``Price("10 HIVE", "1 HBD")``
    * ``Price(Amount("10 HIVE"), Amount("1 HBD"))``
    * ``Price(1.0, "HBD/HIVE")``

    Instances of this class can be used in regular mathematical expressions
    (``+-*/%``) such as:

    .. code-block:: python

        >>> from nectar.price import Price
        >>> from nectar import Hive
        >>> hv = Hive("https://api.hive.blog")
        >>> Price("0.3314 HBD/HIVE", blockchain_instance=hv) * 2
        0.662804 HBD/HIVE
        >>> Price(0.3314, "HBD", "HIVE", blockchain_instance=hv)
        0.331402 HBD/HIVE

    """

    def __init__(
        self,
        price=None,
        base=None,
        quote=None,
        base_asset=None,  # to identify sell/buy
        blockchain_instance=None,
    ):
        """
        Initialize a Price object representing a ratio between a base and quote asset.

        This constructor accepts multiple input forms and normalizes them into internal
        "base" and "quote" Amount entries. Supported usages:
        - price: str like "X BASE/QUOTE" with no base/quote: parses symbols and creates
          Amounts from the fractional representation of X.
        - price: dict with "base" and "quote": loads Amounts directly (raises AssertionError
          if a top-level "price" key is present).
        - price: numeric (float/int/Decimal) with base and quote provided as Asset or
          symbol strings: converts the numeric value to a Fraction and builds Amounts.
        - price: str representing an Amount and base: when price is a string and base is
          a symbol string, price and base are used to build quote/base Amounts.
        - price and base as Amount instances: accepts Amount objects directly.
        - price is None with base and quote as symbol strings or Amounts: loads assets
          or Amounts respectively.

        Parameters (not exhaustive):
        - price: numeric, str, dict, or Amount — the price or a representation used to
          derive base/quote Amounts.
        - base: Asset, Amount, or str — identifies the base side (or a symbol string
          used to parse both symbols when combined with a numeric price).
        - quote: Asset, Amount, or str — identifies the quote side.
        - base_asset: optional; used only as an identifier flag for buy/sell contexts.
        - blockchain_instance: blockchain context used to construct Asset/Amount (omitted
          from param listing as a shared service).

        Raises:
        - AssertionError: if a dict `price` includes a top-level "price" key.
        - ValueError: if the combination of inputs cannot be parsed into base and quote.
        """
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        if price == "":
            price = None
        if price is not None and isinstance(price, str) and not base and not quote:
            price, assets = price.split(" ")
            base_symbol, quote_symbol = assets_from_string(assets)
            base = Asset(base_symbol, blockchain_instance=self.blockchain)
            quote = Asset(quote_symbol, blockchain_instance=self.blockchain)
            frac = Fraction(float(price)).limit_denominator(10 ** base["precision"])
            self["quote"] = Amount(
                amount=frac.denominator, asset=quote, blockchain_instance=self.blockchain
            )
            self["base"] = Amount(
                amount=frac.numerator, asset=base, blockchain_instance=self.blockchain
            )

        elif price is not None and isinstance(price, dict) and "base" in price and "quote" in price:
            if "price" in price:
                raise AssertionError("You cannot provide a 'price' this way")
            # Regular 'price' objects according to hive-core
            # base_id = price["base"]["asset_id"]
            # if price["base"]["asset_id"] == base_id:
            self["base"] = Amount(price["base"], blockchain_instance=self.blockchain)
            self["quote"] = Amount(price["quote"], blockchain_instance=self.blockchain)
            # else:
            #    self["quote"] = Amount(price["base"], blockchain_instance=self.blockchain)
            #    self["base"] = Amount(price["quote"], blockchain_instance=self.blockchain)

        elif price is not None and isinstance(base, Asset) and isinstance(quote, Asset):
            frac = Fraction(float(price)).limit_denominator(10 ** base["precision"])
            self["quote"] = Amount(
                amount=frac.denominator, asset=quote, blockchain_instance=self.blockchain
            )
            self["base"] = Amount(
                amount=frac.numerator, asset=base, blockchain_instance=self.blockchain
            )

        elif price is not None and isinstance(base, str) and isinstance(quote, str):
            base = Asset(base, blockchain_instance=self.blockchain)
            quote = Asset(quote, blockchain_instance=self.blockchain)
            frac = Fraction(float(price)).limit_denominator(10 ** base["precision"])
            self["quote"] = Amount(
                amount=frac.denominator, asset=quote, blockchain_instance=self.blockchain
            )
            self["base"] = Amount(
                amount=frac.numerator, asset=base, blockchain_instance=self.blockchain
            )

        elif price is None and isinstance(base, str) and isinstance(quote, str):
            self["quote"] = Amount(quote, blockchain_instance=self.blockchain)
            self["base"] = Amount(base, blockchain_instance=self.blockchain)
        elif price is not None and isinstance(price, str) and isinstance(base, str):
            self["quote"] = Amount(price, blockchain_instance=self.blockchain)
            self["base"] = Amount(base, blockchain_instance=self.blockchain)
        # len(args) > 1

        elif isinstance(price, Amount) and isinstance(base, Amount):
            self["quote"], self["base"] = price, base

        # len(args) == 0
        elif price is None and isinstance(base, Amount) and isinstance(quote, Amount):
            self["quote"] = quote
            self["base"] = base

        elif (
            isinstance(price, float) or isinstance(price, int) or isinstance(price, Decimal)
        ) and isinstance(base, str):
            base_symbol, quote_symbol = assets_from_string(base)
            base = Asset(base_symbol, blockchain_instance=self.blockchain)
            quote = Asset(quote_symbol, blockchain_instance=self.blockchain)
            frac = Fraction(float(price)).limit_denominator(10 ** base["precision"])
            self["quote"] = Amount(
                amount=frac.denominator, asset=quote, blockchain_instance=self.blockchain
            )
            self["base"] = Amount(
                amount=frac.numerator, asset=base, blockchain_instance=self.blockchain
            )

        else:
            raise ValueError("Couldn't parse 'Price'.")

    def __setitem__(self, key, value):
        dict.__setitem__(self, key, value)
        if (
            "quote" in self and "base" in self and self["base"] and self["quote"]
        ):  # don't derive price for deleted Orders
            dict.__setitem__(
                self, "price", self._safedivide(self["base"]["amount"], self["quote"]["amount"])
            )

    def copy(self):
        return Price(
            None,
            base=self["base"].copy(),
            quote=self["quote"].copy(),
            blockchain_instance=self.blockchain,
        )

    def _safedivide(self, a, b):
        if b != 0.0:
            return a / b
        else:
            return float("Inf")

    def symbols(self):
        return self["base"]["symbol"], self["quote"]["symbol"]

    def as_base(self, base):
        """
        Return a copy of this Price expressed with the given asset as the base.

        If `base` matches the current base symbol this returns a shallow copy.
        If `base` matches the current quote symbol this returns a copy with base and quote inverted.
        Raises InvalidAssetException if `base` is neither the base nor the quote of this price.

        Parameters:
            base (str): Asset symbol to use as the base (e.g., "HIVE" or "HBD").

        Returns:
            Price: A new Price instance whose base asset is `base`.
        """
        if base == self["base"]["symbol"]:
            return self.copy()
        elif base == self["quote"]["symbol"]:
            return self.copy().invert()
        else:
            raise InvalidAssetException

    def as_quote(self, quote):
        """
        Return a Price instance expressed with the given quote asset symbol.

        If `quote` matches the current quote symbol, returns a copy of this Price.
        If `quote` matches the current base symbol, returns a copied, inverted Price.
        A new object is always returned (the original is not modified).

        Parameters:
            quote (str): Asset symbol to use as the quote (e.g., "HBD" or "HIVE").

        Returns:
            Price: A Price object with `quote` as the quote asset.

        Raises:
            InvalidAssetException: If `quote` does not match either the current base or quote symbol.
        """
        if quote == self["quote"]["symbol"]:
            return self.copy()
        elif quote == self["base"]["symbol"]:
            return self.copy().invert()
        else:
            raise InvalidAssetException

    def invert(self):
        """
        Invert the price in place, swapping base and quote assets (e.g., HBD/HIVE -> HIVE/HBD).

        Returns:
            self: The same Price instance after inversion.

        Example:
            >>> from nectar.price import Price
            >>> from nectar import Hive
            >>> hv = Hive("https://api.hive.blog")
            >>> Price("0.3314 HBD/HIVE", blockchain_instance=hv).invert()
            3.017483 HIVE/HBD
        """
        tmp = self["quote"]
        self["quote"] = self["base"]
        self["base"] = tmp
        return self

    def json(self):
        return {"base": self["base"].json(), "quote": self["quote"].json()}

    def __repr__(self):
        return "{price:.{precision}f} {base}/{quote}".format(
            price=self["price"],
            base=self["base"]["symbol"],
            quote=self["quote"]["symbol"],
            precision=(self["base"]["asset"]["precision"] + self["quote"]["asset"]["precision"]),
        )

    def __float__(self):
        return float(self["price"])

    def _check_other(self, other):
        if not other["base"]["symbol"] == self["base"]["symbol"]:
            raise AssertionError()
        if not other["quote"]["symbol"] == self["quote"]["symbol"]:
            raise AssertionError()

    def __mul__(self, other):
        a = self.copy()
        if isinstance(other, Price):
            # Rotate/invert other
            if (
                self["quote"]["symbol"] not in other.symbols()
                and self["base"]["symbol"] not in other.symbols()
            ):
                raise InvalidAssetException

            # base/quote = a/b
            # a/b * b/c = a/c
            a = self.copy()
            if self["quote"]["symbol"] == other["base"]["symbol"]:
                a["base"] = Amount(
                    float(self["base"]) * float(other["base"]),
                    self["base"]["symbol"],
                    blockchain_instance=self.blockchain,
                )
                a["quote"] = Amount(
                    float(self["quote"]) * float(other["quote"]),
                    other["quote"]["symbol"],
                    blockchain_instance=self.blockchain,
                )
            # a/b * c/a =  c/b
            elif self["base"]["symbol"] == other["quote"]["symbol"]:
                a["base"] = Amount(
                    float(self["base"]) * float(other["base"]),
                    other["base"]["symbol"],
                    blockchain_instance=self.blockchain,
                )
                a["quote"] = Amount(
                    float(self["quote"]) * float(other["quote"]),
                    self["quote"]["symbol"],
                    blockchain_instance=self.blockchain,
                )
            else:
                raise ValueError("Wrong rotation of prices")
        elif isinstance(other, Amount):
            check_asset(other["asset"], self["quote"]["asset"], self.blockchain)
            a = other.copy() * self["price"]
            a["asset"] = self["base"]["asset"].copy()
            a["symbol"] = self["base"]["asset"]["symbol"]
        else:
            a["base"] *= other
        return a

    def __imul__(self, other):
        if isinstance(other, Price):
            tmp = self * other
            self["base"] = tmp["base"]
            self["quote"] = tmp["quote"]
        else:
            self["base"] *= other
        return self

    def __div__(self, other):
        a = self.copy()
        if isinstance(other, Price):
            # Rotate/invert other
            if sorted(self.symbols()) == sorted(other.symbols()):
                return float(self.as_base(self["base"]["symbol"])) / float(
                    other.as_base(self["base"]["symbol"])
                )
            elif self["quote"]["symbol"] in other.symbols():
                other = other.as_base(self["quote"]["symbol"])
            elif self["base"]["symbol"] in other.symbols():
                other = other.as_base(self["base"]["symbol"])
            else:
                raise InvalidAssetException
            a["base"] = Amount(
                float(self["base"].amount / other["base"].amount),
                other["quote"]["symbol"],
                blockchain_instance=self.blockchain,
            )
            a["quote"] = Amount(
                float(self["quote"].amount / other["quote"].amount),
                self["quote"]["symbol"],
                blockchain_instance=self.blockchain,
            )
        elif isinstance(other, Amount):
            check_asset(other["asset"], self["quote"]["asset"], self.blockchain)
            a = other.copy() / self["price"]
            a["asset"] = self["base"]["asset"].copy()
            a["symbol"] = self["base"]["asset"]["symbol"]
        else:
            a["base"] /= other
        return a

    def __idiv__(self, other):
        if isinstance(other, Price):
            tmp = self / other
            self["base"] = tmp["base"]
            self["quote"] = tmp["quote"]
        else:
            self["base"] /= other
        return self

    def __floordiv__(self, other):
        raise NotImplementedError("This is not possible as the price is a ratio")

    def __ifloordiv__(self, other):
        raise NotImplementedError("This is not possible as the price is a ratio")

    def __lt__(self, other):
        if isinstance(other, Price):
            self._check_other(other)
            return self["price"] < other["price"]
        else:
            return self["price"] < float(other or 0)

    def __le__(self, other):
        if isinstance(other, Price):
            self._check_other(other)
            return self["price"] <= other["price"]
        else:
            return self["price"] <= float(other or 0)

    def __eq__(self, other):
        if isinstance(other, Price):
            self._check_other(other)
            return self["price"] == other["price"]
        else:
            return self["price"] == float(other or 0)

    def __ne__(self, other):
        if isinstance(other, Price):
            self._check_other(other)
            return self["price"] != other["price"]
        else:
            return self["price"] != float(other or 0)

    def __ge__(self, other):
        if isinstance(other, Price):
            self._check_other(other)
            return self["price"] >= other["price"]
        else:
            return self["price"] >= float(other or 0)

    def __gt__(self, other):
        if isinstance(other, Price):
            self._check_other(other)
            return self["price"] > other["price"]
        else:
            return self["price"] > float(other or 0)

    __truediv__ = __div__
    __truemul__ = __mul__
    __str__ = __repr__

    @property
    def market(self):
        """Open the corresponding market

        :returns: Instance of :class:`nectar.market.Market` for the
                  corresponding pair of assets.
        """
        from .market import Market

        return Market(
            base=self["base"]["asset"],
            quote=self["quote"]["asset"],
            blockchain_instance=self.blockchain,
        )


class Order(Price):
    """This class inherits :class:`nectar.price.Price` but has the ``base``
    and ``quote`` Amounts not only be used to represent the price (as a
    ratio of base and quote) but instead has those amounts represent the
    amounts of an actual order!

    :param Hive blockchain_instance: Hive instance

    .. note::

            If an order is marked as deleted, it will carry the
            'deleted' key which is set to ``True`` and all other
            data be ``None``.
    """

    def __init__(self, base, quote=None, blockchain_instance=None, **kwargs):
        self.blockchain = blockchain_instance or shared_blockchain_instance()

        if isinstance(base, dict) and "sell_price" in base:
            super(Order, self).__init__(base["sell_price"], blockchain_instance=self.blockchain)
            self["id"] = base.get("id")
        elif isinstance(base, dict) and "min_to_receive" in base and "amount_to_sell" in base:
            super(Order, self).__init__(
                Amount(base["min_to_receive"], blockchain_instance=self.blockchain),
                Amount(base["amount_to_sell"], blockchain_instance=self.blockchain),
                blockchain_instance=self.blockchain,
            )
            self["id"] = base.get("id")
        elif isinstance(base, Amount) and isinstance(quote, Amount):
            super(Order, self).__init__(
                None, base=base, quote=quote, blockchain_instance=self.blockchain
            )
        else:
            raise ValueError("Unknown format to load Order")

    def __repr__(self):
        if "deleted" in self and self["deleted"]:
            return "deleted order %s" % self["id"]
        else:
            t = ""
            if "time" in self and self["time"]:
                t += "(%s) " % self["time"]
            if "type" in self and self["type"]:
                t += "%s " % str(self["type"])
            if "quote" in self and self["quote"]:
                t += "%s " % str(self["quote"])
            if "base" in self and self["base"]:
                t += "%s " % str(self["base"])
            return t + "@ " + Price.__repr__(self)

    __str__ = __repr__


class FilledOrder(Price):
    """This class inherits :class:`nectar.price.Price` but has the ``base``
    and ``quote`` Amounts not only be used to represent the price (as a
    ratio of base and quote) but instead has those amounts represent the
    amounts of an actually filled order!

    :param Hive blockchain_instance: Hive instance

    .. note:: Instances of this class come with an additional ``date`` key
              that shows when the order has been filled!
    """

    def __init__(self, order, blockchain_instance=None, **kwargs):
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        if isinstance(order, dict) and "current_pays" in order and "open_pays" in order:
            # filled orders from account history
            if "op" in order:
                order = order["op"]

            super(FilledOrder, self).__init__(
                Amount(order["open_pays"], blockchain_instance=self.blockchain),
                Amount(order["current_pays"], blockchain_instance=self.blockchain),
                blockchain_instance=self.blockchain,
            )
            if "date" in order:
                self["date"] = formatTimeString(order["date"])

        else:
            raise ValueError("Couldn't parse 'Price'.")

    def json(self):
        return {
            "date": formatTimeString(self["date"]),
            "current_pays": self["base"].json(),
            "open_pays": self["quote"].json(),
        }

    def __repr__(self):
        t = ""
        if "date" in self and self["date"]:
            t += "(%s) " % self["date"]
        if "type" in self and self["type"]:
            t += "%s " % str(self["type"])
        if "quote" in self and self["quote"]:
            t += "%s " % str(self["quote"])
        if "base" in self and self["base"]:
            t += "%s " % str(self["base"])
        return t + "@ " + Price.__repr__(self)

    __str__ = __repr__



import argparse
import logging
import sys
from datetime import timedelta
from timeit import default_timer as timer

from prettytable import PrettyTable

from nectar import Hive
from nectar.block import Block
from nectar.blockchain import Blockchain
from nectar.nodelist import NodeList
from nectar.utils import parse_time

log = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


def parse_args(args=None):
    """
    Parse command-line arguments for the script.

    One-line summary:
        Parse CLI arguments and return the parsed Namespace.

    Detailed:
        Accepts an optional list of argument strings (for testing). Parses a single optional
        positional argument "blockchain" which is a string specifying the blockchain to use;
        if omitted the parser will use standard input as the default value.

    Parameters:
        args (list[str] | None): Optional list of argument strings to parse. If None, uses
            arguments from the environment (sys.argv).

    Returns:
        argparse.Namespace: Parsed arguments with attribute `blockchain`.
    """
    d = "Show op type stats for hive."
    parser = argparse.ArgumentParser(description=d)
    parser.add_argument(
        "blockchain",
        type=str,
        nargs="?",
        default=sys.stdin,
        help="Blockchain (hive)",
    )
    return parser.parse_args(args)


def main(args=None):
    """
    Compute and print operation-type statistics for recent Hive blocks.

    Scans approximately one hour of recent Hive blocks (by default) and tallies operation types across all transactions found in that window. Results are printed as a table showing each operation type, its count, and its percentage of total operations.

    Parameters:
        args (list|None): Optional argument list in the same format as sys.argv[1:] or None to use command-line arguments parsed by argparse. Recognized options include a single optional positional "blockchain" identifier (defaults to Hive).
    """
    args = parse_args(args)
    blockchain = args.blockchain

    nodelist = NodeList()
    nodelist.update_nodes(weights={"block": 1})

    max_batch_size = 50
    threading = False
    thread_num = 16
    block_debug = 1000

    nodes = nodelist.get_hive_nodes()
    blk_inst = Hive(node=nodes, num_retries=3, num_retries_call=3, timeout=30)

    print(blk_inst)
    block_count = 0
    total_ops = 0
    total_trx = 0
    duration_s = 60 * 60 * 1
    blocksperday = int(duration_s / 3)

    blockchain = Blockchain(
        blockchain_instance=blk_inst,
    )
    current_block_num = blockchain.get_current_block_num()
    last_block_id = current_block_num - blocksperday

    last_block = Block(last_block_id, blockchain_instance=blk_inst)

    stopTime = last_block.time() + timedelta(seconds=duration_s)

    start = timer()
    op_stats = {}
    for entry in blockchain.blocks(
        start=last_block_id,
        max_batch_size=max_batch_size,
        threading=threading,
        thread_num=thread_num,
    ):
        if "block" in entry:
            block_time = parse_time(entry["block"]["timestamp"])
        else:
            block_time = entry["timestamp"]
        if block_time > stopTime:
            break
        block_count += 1
        if "block" in entry:
            trxs = entry["block"]["transactions"]
        else:
            trxs = entry["transactions"]
        for tx in trxs:
            total_trx += 1
            for op in tx["operations"]:
                if "_operation" in op["type"]:
                    op_type = op["type"][:-10]
                else:
                    op_type = op["type"]
                if op_type in op_stats:
                    op_stats[op_type] += 1
                else:
                    op_stats[op_type] = 1
                total_ops += 1

        ops_per_day = total_ops / block_count * blocksperday
        if block_count % (block_debug) == 0:
            print(
                "%d blocks remaining... estimated ops per day: %.1f"
                % (blocksperday - block_count, ops_per_day)
            )

    _duration = timer() - start
    t = PrettyTable(["Type", "Count", "percentage"])
    t.align = "l"
    op_list = []
    for o in op_stats:
        op_list.append({"type": o, "n": op_stats[o], "perc": op_stats[o] / total_ops * 100})
    op_list_sorted = sorted(op_list, key=lambda x: x["n"], reverse=True)
    for op in op_list_sorted:
        t.add_row([op["type"], op["n"], "%.2f %%" % op["perc"]])
    print(t)


if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-
import json

import pytest

from nectar.haf import HAF


class FakeResponse:
    def __init__(self, status_code=200, payload=None, raise_for_status_exc=None, bad_json=False):
        self.status_code = status_code
        self._payload = payload if payload is not None else {}
        self._raise_exc = raise_for_status_exc
        self._bad_json = bad_json
        self.text = json.dumps(self._payload) if not bad_json else "<html>not json</html>"

    def raise_for_status(self):
        if self._raise_exc:
            raise self._raise_exc

    def json(self):
        if self._bad_json:
            # Simulate invalid JSON parse error
            raise json.JSONDecodeError("Expecting value", self.text, 0)
        return self._payload


@pytest.fixture
def patch_requests(monkeypatch):
    calls = {
        "last": None,
        "history": [],
    }

    def _request(method, url, headers=None, **kwargs):
        # Default success response returning an echo of the endpoint
        payload = kwargs.pop("_payload", {"ok": True, "url": url, "method": method})
        resp = FakeResponse(payload=payload)
        calls["last"] = (method, url, headers, kwargs)
        calls["history"].append((method, url))
        return resp

    monkeypatch.setattr("requests.request", _request)
    return calls


def test_init_and_current_api_default():
    haf = HAF()
    assert haf.get_current_api() in HAF.DEFAULT_APIS


def test_init_invalid_api_raises():
    with pytest.raises(ValueError):
        HAF(api="invalid://nope")


def test_set_api_switch_and_strip():
    haf = HAF()
    haf.set_api("https://api.syncad.com/")
    assert haf.get_current_api() == "https://api.syncad.com"


def test_get_available_apis():
    haf = HAF()
    apis = haf.get_available_apis()
    assert isinstance(apis, list)
    assert all(api.startswith("http") for api in apis)


def test_reputation_success(monkeypatch):
    # Arrange: respond with a simple reputation payload
    def responder(method, url, headers=None, **kwargs):
        return FakeResponse(payload={"account": "alice", "reputation": 70})

    monkeypatch.setattr("requests.request", responder)

    haf = HAF()
    data = haf.reputation("alice")
    assert data is not None
    assert data["account"] == "alice"
    assert "reputation" in data

def test_reputation_invalid_account_raises_value_error():
    haf = HAF()
    with pytest.raises(ValueError):
        haf.reputation(123)  # type: ignore[arg-type]
    with pytest.raises(ValueError):
        haf.reputation("")


@pytest.mark.parametrize(
    "method_name, endpoint_suffix, sample_payload",
    [
        (
            "get_account_balances",
            "balance-api/accounts/alice/balances",
            {"hive_balance": "1.000 HIVE"},
        ),
        (
            "get_account_delegations",
            "balance-api/accounts/alice/delegations",
            {"incoming_delegations": []},
        ),
        (
            "get_account_recurrent_transfers",
            "balance-api/accounts/alice/recurrent-transfers",
            {"outgoing_recurrent_transfers": []},
        ),
        ("get_reputation_version", "reputation-api/version", "1.0.0"),
        ("get_reputation_last_synced_block", "reputation-api/last-synced-block", 12345678),
        ("get_balance_version", "balance-api/version", "2.0.0"),
        ("get_balance_last_synced_block", "balance-api/last-synced-block", 87654321),
    ],
)
def test_haf_methods_success(monkeypatch, method_name, endpoint_suffix, sample_payload):
    def _request(method, url, headers=None, **kwargs):
        # Verify we hit the expected endpoint
        assert url.endswith("/" + endpoint_suffix)
        return FakeResponse(payload=sample_payload)

    monkeypatch.setattr("requests.request", _request)

    haf = HAF()
    method = getattr(haf, method_name)
    # Method may require account arg
    if "accounts/alice" in endpoint_suffix:
        result = method("alice")
    else:
        result = method()
    assert result == sample_payload


@pytest.mark.parametrize(
    "method_name, needs_account",
    [
        ("reputation", True),
        ("get_account_balances", True),
        ("get_account_delegations", True),
        ("get_account_recurrent_transfers", True),
        ("get_reputation_version", False),
        ("get_reputation_last_synced_block", False),
        ("get_balance_version", False),
        ("get_balance_last_synced_block", False),
    ],
)
def test_haf_methods_request_exception_returns_none(monkeypatch, method_name, needs_account):
    class Boom(Exception):
        pass

    def _request(method, url, headers=None, **kwargs):
        raise Boom("network down")

    monkeypatch.setattr("requests.request", _request)

    haf = HAF()
    method = getattr(haf, method_name)
    result = method("alice") if needs_account else method()
    assert result is None


@pytest.mark.parametrize(
    "method_name, needs_account",
    [
        ("reputation", True),
        ("get_account_balances", True),
        ("get_account_delegations", True),
        ("get_account_recurrent_transfers", True),
        ("get_reputation_version", False),
        ("get_reputation_last_synced_block", False),
        ("get_balance_version", False),
        ("get_balance_last_synced_block", False),
    ],
)
def test_haf_methods_invalid_json_returns_none(monkeypatch, method_name, needs_account):
    def _request(method, url, headers=None, **kwargs):
        # Valid HTTP, but broken JSON
        return FakeResponse(payload=None, bad_json=True)

    monkeypatch.setattr("requests.request", _request)

    haf = HAF()
    method = getattr(haf, method_name)
    result = method("alice") if needs_account else method()
    assert result is None



# -*- coding: utf-8 -*-
import json

try:
    from collections.abc import Mapping  # noqa
except ImportError:
    from collections import Mapping  # noqa


class DotDict(dict):
    def __init__(self, *args):
        """This class simplifies the use of "."-separated
        keys when defining a nested dictionary:::

            >>> from nectar.profile import Profile
            >>> keys = ['profile.url', 'profile.img']
            >>> values = ["http:", "foobar"]
            >>> p = Profile(keys, values)
            >>> print(p["profile"]["url"])
            http:

        """
        if len(args) == 2:
            for i, item in enumerate(args[0]):
                t = self
                parts = item.split(".")
                for j, part in enumerate(parts):
                    if j < len(parts) - 1:
                        t = t.setdefault(part, {})
                    else:
                        t[part] = args[1][i]
        elif len(args) == 1 and isinstance(args[0], dict):
            for k, v in args[0].items():
                self[k] = v
        elif len(args) == 1 and isinstance(args[0], str):
            for k, v in json.loads(args[0]).items():
                self[k] = v


class Profile(DotDict):
    """This class is a template to model a user's on-chain
    profile according to Hive profile metadata conventions.
    """

    def __init__(self, *args, **kwargs):
        """
        Initialize a Profile by delegating to the DotDict initializer.

        This constructor accepts the same arguments as DotDict (e.g., dot-separated key/value pairs, a dict, or a JSON string) and performs no additional processing.
        """
        super(Profile, self).__init__(*args, **kwargs)

    def __str__(self):
        return json.dumps(self)

    def update(self, u):
        for k, v in u.items():
            if isinstance(v, Mapping):
                self.setdefault(k, {}).update(v)
            else:
                self[k] = u[k]

    def remove(self, key):
        parts = key.split(".")
        if len(parts) > 1:
            self[parts[0]].pop(".".join(parts[1:]))
        else:
            super(Profile, self).pop(parts[0], None)



# -*- coding: utf-8 -*-


class NectarException(Exception):
    """Base exception for all Nectar-related errors"""

    pass


class WalletExists(NectarException):
    """A wallet has already been created and requires a password to be
    unlocked by means of :func:`nectar.wallet.Wallet.unlock`.
    """

    pass


class RPCConnectionRequired(NectarException):
    """An RPC connection is required"""

    pass


class InvalidMemoKeyException(NectarException):
    """Memo key in message is invalid"""

    pass


class WrongMemoKey(NectarException):
    """The memo provided is not equal the one on the blockchain"""

    pass


class OfflineHasNoRPCException(NectarException):
    """When in offline mode, we don't have RPC"""

    pass


class AccountExistsException(NectarException):
    """The requested account already exists"""

    pass


class AccountDoesNotExistsException(NectarException):
    """The account does not exist"""

    pass


class AssetDoesNotExistsException(NectarException):
    """The asset does not exist"""

    pass


class InvalidAssetException(NectarException):
    """An invalid asset has been provided"""

    pass


class InsufficientAuthorityError(NectarException):
    """The transaction requires signature of a higher authority"""

    pass


class VotingInvalidOnArchivedPost(NectarException):
    """The transaction requires signature of a higher authority"""

    pass


class MissingKeyError(NectarException):
    """A required key couldn't be found in the wallet"""

    pass


class InvalidWifError(NectarException):
    """The provided private Key has an invalid format"""

    pass


class BlockDoesNotExistsException(NectarException):
    """The block does not exist"""

    pass


class NoWalletException(NectarException):
    """No Wallet could be found, please use :func:`nectar.wallet.Wallet.create` to
    create a new wallet
    """

    pass


class WitnessDoesNotExistsException(NectarException):
    """The witness does not exist"""

    pass


class ContentDoesNotExistsException(NectarException):
    """The content does not exist"""

    pass


class VoteDoesNotExistsException(NectarException):
    """The vote does not exist"""

    pass


class WrongMasterPasswordException(NectarException):
    """The password provided could not properly unlock the wallet"""

    pass


class VestingBalanceDoesNotExistsException(NectarException):
    """Vesting Balance does not exist"""

    pass


class InvalidMessageSignature(NectarException):
    """The message signature does not fit the message"""

    pass


class NoWriteAccess(NectarException):
    """Cannot store to sqlite3 database due to missing write access"""

    pass


class BatchedCallsNotSupported(NectarException):
    """Batch calls do not work"""

    pass


class BlockWaitTimeExceeded(NectarException):
    """Wait time for new block exceeded"""

    pass



# This Python file uses the following encoding: utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import random
import sys
import unittest
from binascii import hexlify, unhexlify
from builtins import str

from nectargraphenebase.account import (
    Address,
    BitcoinAddress,
    BrainKey,
    Mnemonic,
    MnemonicKey,
    PasswordKey,
    PrivateKey,
    PublicKey,
)
from nectargraphenebase.base58 import Base58
from nectargraphenebase.bip32 import BIP32Key


class Testcases(unittest.TestCase):
    def test_B85hexgetb58_btc(self):
        self.assertEqual(
            [
                "5HqUkGuo62BfcJU5vNhTXKJRXuUi9QSE6jp8C3uBJ2BVHtB8WSd",
                "5JWcdkhL3w4RkVPcZMdJsjos22yB5cSkPExerktvKnRNZR5gx1S",
                "5HvVz6XMx84aC5KaaBbwYrRLvWE46cH6zVnv4827SBPLorg76oq",
                "5Jete5oFNjjk3aUMkKuxgAXsp7ZyhgJbYNiNjHLvq5xzXkiqw7R",
                "5KDT58ksNsVKjYShG4Ls5ZtredybSxzmKec8juj7CojZj6LPRF7",
                "02b52e04a0acfe611a4b6963462aca94b6ae02b24e321eda86507661901adb49",
                "5b921f7051be5e13e177a0253229903c40493df410ae04f4a450c85568f19131",
                "0e1bfc9024d1f55a7855dc690e45b2e089d2d825a4671a3c3c7e4ea4e74ec00e",
                "6e5cc4653d46e690c709ed9e0570a2c75a286ad7c1bc69a648aae6855d919d3e",
            ],
            [
                format(
                    Base58("02b52e04a0acfe611a4b6963462aca94b6ae02b24e321eda86507661901adb49"),
                    "WIF",
                ),
                format(
                    Base58("5b921f7051be5e13e177a0253229903c40493df410ae04f4a450c85568f19131"),
                    "WIF",
                ),
                format(
                    Base58("0e1bfc9024d1f55a7855dc690e45b2e089d2d825a4671a3c3c7e4ea4e74ec00e"),
                    "WIF",
                ),
                format(
                    Base58("6e5cc4653d46e690c709ed9e0570a2c75a286ad7c1bc69a648aae6855d919d3e"),
                    "WIF",
                ),
                format(
                    Base58("b84abd64d66ee1dd614230ebbe9d9c6d66d78d93927c395196666762e9ad69d8"),
                    "WIF",
                ),
                repr(Base58("5HqUkGuo62BfcJU5vNhTXKJRXuUi9QSE6jp8C3uBJ2BVHtB8WSd")),
                repr(Base58("5JWcdkhL3w4RkVPcZMdJsjos22yB5cSkPExerktvKnRNZR5gx1S")),
                repr(Base58("5HvVz6XMx84aC5KaaBbwYrRLvWE46cH6zVnv4827SBPLorg76oq")),
                repr(Base58("5Jete5oFNjjk3aUMkKuxgAXsp7ZyhgJbYNiNjHLvq5xzXkiqw7R")),
            ],
        )

    def test_B85hexgetb58(self):
        self.assertEqual(
            [
                "BTS2CAbTi1ZcgMJ5otBFZSGZJKJenwGa9NvkLxsrS49Kr8JsiSGc",
                "BTShL45FEyUVSVV1LXABQnh4joS9FsUaffRtsdarB5uZjPsrwMZF",
                "BTS7DQR5GsfVaw4wJXzA3TogDhuQ8tUR2Ggj8pwyNCJXheHehL4Q",
                "BTSqc4QMAJHAkna65i8U4b7nkbWk4VYSWpZebW7JBbD7MN8FB5sc",
                "BTS2QAVTJnJQvLUY4RDrtxzX9jS39gEq8gbqYMWjgMxvsvZTJxDSu",
            ],
            [
                format(
                    Base58("02b52e04a0acfe611a4b6963462aca94b6ae02b24e321eda86507661901adb49"),
                    "BTS",
                ),
                format(
                    Base58("5b921f7051be5e13e177a0253229903c40493df410ae04f4a450c85568f19131"),
                    "BTS",
                ),
                format(
                    Base58("0e1bfc9024d1f55a7855dc690e45b2e089d2d825a4671a3c3c7e4ea4e74ec00e"),
                    "BTS",
                ),
                format(
                    Base58("6e5cc4653d46e690c709ed9e0570a2c75a286ad7c1bc69a648aae6855d919d3e"),
                    "BTS",
                ),
                format(
                    Base58("b84abd64d66ee1dd614230ebbe9d9c6d66d78d93927c395196666762e9ad69d8"),
                    "BTS",
                ),
            ],
        )

    def test_Address(self):
        self.assertEqual(
            [
                format(Address("BTSFN9r6VYzBK8EKtMewfNbfiGCr56pHDBFi", prefix="BTS"), "BTS"),
                format(Address("BTSdXrrTXimLb6TEt3nHnePwFmBT6Cck112", prefix="BTS"), "BTS"),
                format(Address("BTSJQUAt4gz4civ8gSs5srTK4r82F7HvpChk", prefix="BTS"), "BTS"),
                format(Address("BTSFPXXHXXGbyTBwdKoJaAPXRnhFNtTRS4EL", prefix="BTS"), "BTS"),
                format(Address("BTS3qXyZnjJneeAddgNDYNYXbF7ARZrRv5dr", prefix="BTS"), "BTS"),
            ],
            [
                "BTSFN9r6VYzBK8EKtMewfNbfiGCr56pHDBFi",
                "BTSdXrrTXimLb6TEt3nHnePwFmBT6Cck112",
                "BTSJQUAt4gz4civ8gSs5srTK4r82F7HvpChk",
                "BTSFPXXHXXGbyTBwdKoJaAPXRnhFNtTRS4EL",
                "BTS3qXyZnjJneeAddgNDYNYXbF7ARZrRv5dr",
            ],
        )

    def test_PubKey(self):
        self.assertEqual(
            [
                format(
                    PublicKey(
                        "BTS6UtYWWs3rkZGV8JA86qrgkG6tyFksgECefKE1MiH4HkLD8PFGL", prefix="BTS"
                    ).address,
                    "BTS",
                ),
                format(
                    PublicKey(
                        "BTS8YAMLtNcnqGNd3fx28NP3WoyuqNtzxXpwXTkZjbfe9scBmSyGT", prefix="BTS"
                    ).address,
                    "BTS",
                ),
                format(
                    PublicKey(
                        "BTS7HUo6bm7Gfoi3RqAtzwZ83BFCwiCZ4tp37oZjtWxGEBJVzVVGw", prefix="BTS"
                    ).address,
                    "BTS",
                ),
                format(
                    PublicKey(
                        "BTS6676cZ9qmqPnWMrm4McjCuHcnt6QW5d8oRJ4t8EDH8DdCjvh4V", prefix="BTS"
                    ).address,
                    "BTS",
                ),
                format(
                    PublicKey(
                        "BTS7u8m6zUNuzPNK1tPPLtnipxgqV9mVmTzrFNJ9GvovvSTCkVUra", prefix="BTS"
                    ).address,
                    "BTS",
                ),
            ],
            [
                "BTS66FCjYKzMwLbE3a59YpmFqA9bwporT4L3",
                "BTSKNpRuPX8KhTBsJoFp1JXd7eQEsnCpRw3k",
                "BTS838ENJargbUrxXWuE2xD9HKjQaS17GdCd",
                "BTSNsrLFWTziSZASnNJjWafFtGBfSu8VG8KU",
                "BTSDjAGuXzk3WXabBEgKKc8NsuQM412boBdR",
            ],
        )

    def test_btsprivkey(self):
        self.assertEqual(
            [
                format(
                    PrivateKey(
                        "5HqUkGuo62BfcJU5vNhTXKJRXuUi9QSE6jp8C3uBJ2BVHtB8WSd"
                    ).compressed.address,
                    "BTS",
                ),
                format(
                    PrivateKey(
                        "5JWcdkhL3w4RkVPcZMdJsjos22yB5cSkPExerktvKnRNZR5gx1S"
                    ).compressed.address,
                    "BTS",
                ),
                format(
                    PrivateKey(
                        "5HvVz6XMx84aC5KaaBbwYrRLvWE46cH6zVnv4827SBPLorg76oq"
                    ).compressed.address,
                    "BTS",
                ),
                format(
                    PrivateKey(
                        "5Jete5oFNjjk3aUMkKuxgAXsp7ZyhgJbYNiNjHLvq5xzXkiqw7R"
                    ).compressed.address,
                    "BTS",
                ),
                format(
                    PrivateKey(
                        "5KDT58ksNsVKjYShG4Ls5ZtredybSxzmKec8juj7CojZj6LPRF7"
                    ).compressed.address,
                    "BTS",
                ),
            ],
            [
                "BTSFN9r6VYzBK8EKtMewfNbfiGCr56pHDBFi",
                "BTSdXrrTXimLb6TEt3nHnePwFmBT6Cck112",
                "BTSJQUAt4gz4civ8gSs5srTK4r82F7HvpChk",
                "BTSFPXXHXXGbyTBwdKoJaAPXRnhFNtTRS4EL",
                "BTS3qXyZnjJneeAddgNDYNYXbF7ARZrRv5dr",
            ],
        )

    def test_btcprivkey(self):
        self.assertEqual(
            [
                format(
                    PrivateKey(
                        "5HvVz6XMx84aC5KaaBbwYrRLvWE46cH6zVnv4827SBPLorg76oq"
                    ).bitcoin.address,
                    "BTC",
                ),
                format(
                    PrivateKey(
                        "5Jete5oFNjjk3aUMkKuxgAXsp7ZyhgJbYNiNjHLvq5xzXkiqw7R"
                    ).bitcoin.address,
                    "BTC",
                ),
                format(
                    PrivateKey(
                        "5KDT58ksNsVKjYShG4Ls5ZtredybSxzmKec8juj7CojZj6LPRF7"
                    ).bitcoin.address,
                    "BTC",
                ),
            ],
            [
                "1G7qw8FiVfHEFrSt3tDi6YgfAdrDrEM44Z",
                "12c7KAAZfpREaQZuvjC5EhpoN6si9vekqK",
                "1Gu5191CVHmaoU3Zz3prept87jjnpFDrXL",
            ],
        )

    def test_btcprivkeystr(self):
        self.assertEqual(
            [
                str(
                    BitcoinAddress.from_pubkey(
                        PrivateKey("5HvVz6XMx84aC5KaaBbwYrRLvWE46cH6zVnv4827SBPLorg76oq").pubkey
                    )
                ),
                str(
                    BitcoinAddress.from_pubkey(
                        PrivateKey("5Jete5oFNjjk3aUMkKuxgAXsp7ZyhgJbYNiNjHLvq5xzXkiqw7R").pubkey,
                        compressed=True,
                    )
                ),
                str(
                    BitcoinAddress.from_pubkey(
                        PrivateKey("5KDT58ksNsVKjYShG4Ls5ZtredybSxzmKec8juj7CojZj6LPRF7").pubkey,
                        compressed=False,
                    )
                ),
            ],
            [
                "1G7qw8FiVfHEFrSt3tDi6YgfAdrDrEM44Z",
                "1E2jXCkSmLxirL31gHwi1UWTjUBxCgS7pq",
                "1Gu5191CVHmaoU3Zz3prept87jjnpFDrXL",
            ],
        )

    def test_gphprivkeystr(self):
        self.assertEqual(
            [
                str(
                    Address.from_pubkey(
                        PrivateKey("5HvVz6XMx84aC5KaaBbwYrRLvWE46cH6zVnv4827SBPLorg76oq").pubkey
                    )
                ),
                str(
                    Address.from_pubkey(
                        PrivateKey("5Jete5oFNjjk3aUMkKuxgAXsp7ZyhgJbYNiNjHLvq5xzXkiqw7R").pubkey,
                        compressed=True,
                    )
                ),
                str(
                    Address.from_pubkey(
                        PrivateKey("5KDT58ksNsVKjYShG4Ls5ZtredybSxzmKec8juj7CojZj6LPRF7").pubkey,
                        compressed=False,
                        prefix="BTS",
                    )
                ),
            ],
            [
                "STMBXqRucGm7nRkk6jm7BNspTJTWRtNcx7k5",
                "STM5tTDDR6M3mkcyVv16edsw8dGUyNQZrvKU",
                "BTS4XPkBqYw882fH5aR5S8mMKXCaZ1yVA76f",
            ],
        )

    def test_PublicKey(self):
        self.assertEqual(
            [
                str(
                    PublicKey("BTS6UtYWWs3rkZGV8JA86qrgkG6tyFksgECefKE1MiH4HkLD8PFGL", prefix="BTS")
                ),
                str(
                    PublicKey("BTS8YAMLtNcnqGNd3fx28NP3WoyuqNtzxXpwXTkZjbfe9scBmSyGT", prefix="BTS")
                ),
                str(
                    PublicKey("BTS7HUo6bm7Gfoi3RqAtzwZ83BFCwiCZ4tp37oZjtWxGEBJVzVVGw", prefix="BTS")
                ),
                str(
                    PublicKey("BTS6676cZ9qmqPnWMrm4McjCuHcnt6QW5d8oRJ4t8EDH8DdCjvh4V", prefix="BTS")
                ),
                str(
                    PublicKey("BTS7u8m6zUNuzPNK1tPPLtnipxgqV9mVmTzrFNJ9GvovvSTCkVUra", prefix="BTS")
                ),
            ],
            [
                "BTS6UtYWWs3rkZGV8JA86qrgkG6tyFksgECefKE1MiH4HkLD8PFGL",
                "BTS8YAMLtNcnqGNd3fx28NP3WoyuqNtzxXpwXTkZjbfe9scBmSyGT",
                "BTS7HUo6bm7Gfoi3RqAtzwZ83BFCwiCZ4tp37oZjtWxGEBJVzVVGw",
                "BTS6676cZ9qmqPnWMrm4McjCuHcnt6QW5d8oRJ4t8EDH8DdCjvh4V",
                "BTS7u8m6zUNuzPNK1tPPLtnipxgqV9mVmTzrFNJ9GvovvSTCkVUra",
            ],
        )

    def test_Privatekey(self):
        self.assertEqual(
            [
                str(PrivateKey("5HvVz6XMx84aC5KaaBbwYrRLvWE46cH6zVnv4827SBPLorg76oq")),
                str(PrivateKey("5Jete5oFNjjk3aUMkKuxgAXsp7ZyhgJbYNiNjHLvq5xzXkiqw7R")),
                str(PrivateKey("5KDT58ksNsVKjYShG4Ls5ZtredybSxzmKec8juj7CojZj6LPRF7")),
                repr(PrivateKey("5HvVz6XMx84aC5KaaBbwYrRLvWE46cH6zVnv4827SBPLorg76oq")),
                repr(PrivateKey("5Jete5oFNjjk3aUMkKuxgAXsp7ZyhgJbYNiNjHLvq5xzXkiqw7R")),
                repr(PrivateKey("5KDT58ksNsVKjYShG4Ls5ZtredybSxzmKec8juj7CojZj6LPRF7")),
            ],
            [
                "5HvVz6XMx84aC5KaaBbwYrRLvWE46cH6zVnv4827SBPLorg76oq",
                "5Jete5oFNjjk3aUMkKuxgAXsp7ZyhgJbYNiNjHLvq5xzXkiqw7R",
                "5KDT58ksNsVKjYShG4Ls5ZtredybSxzmKec8juj7CojZj6LPRF7",
                "0e1bfc9024d1f55a7855dc690e45b2e089d2d825a4671a3c3c7e4ea4e74ec00e",
                "6e5cc4653d46e690c709ed9e0570a2c75a286ad7c1bc69a648aae6855d919d3e",
                "b84abd64d66ee1dd614230ebbe9d9c6d66d78d93927c395196666762e9ad69d8",
            ],
        )

    def test_BrainKey(self):
        self.assertEqual(
            [
                str(
                    BrainKey(
                        "COLORER BICORN KASBEKE FAERIE LOCHIA GOMUTI SOVKHOZ Y GERMAL AUNTIE PERFUMY TIME FEATURE GANGAN CELEMIN MATZO"
                    ).get_private()
                ),
                str(
                    BrainKey(
                        "NAK TILTING MOOTING TAVERT SCREENY MAGIC BARDIE UPBORNE CONOID MAUVE CARBON NOTAEUM BITUMEN HOOEY KURUMA COWFISH"
                    ).get_private()
                ),
                str(
                    BrainKey(
                        "CORKITE CORDAGE FONDISH UNDER FORGET BEFLEA OUTBUD ZOOGAMY BERLINE ACANTHA STYLO YINCE TROPISM TUNKET FALCULA TOMENT"
                    ).get_private()
                ),
                str(
                    BrainKey(
                        "MURZA PREDRAW FIT LARIGOT CRYOGEN SEVENTH LISP UNTAWED AMBER CRETIN KOVIL TEATED OUTGRIN POTTAGY KLAFTER DABB"
                    ).get_private()
                ),
                str(
                    BrainKey(
                        "VERDICT REPOUR SUNRAY WAMBLY UNFILM UNCOUS COWMAN REBUOY MIURUS KEACORN BENZOLE BEMAUL SAXTIE DOLENT CHABUK BOUGHED"
                    ).get_private()
                ),
                str(
                    BrainKey(
                        "HOUGH TRUMPH SUCKEN EXODY MAMMATE PIGGIN CRIME TEPEE URETHAN TOLUATE BLINDLY CACOEPY SPINOSE COMMIE GRIECE FUNDAL"
                    ).get_private()
                ),
                str(
                    BrainKey(
                        "OERSTED ETHERIN TESTIS PEGGLE ONCOST POMME SUBAH FLOODER OLIGIST ACCUSE UNPLAT OATLIKE DEWTRY CYCLIZE PIMLICO CHICOT"
                    ).get_private()
                ),
            ],
            [
                "5JfwDztjHYDDdKnCpjY6cwUQfM4hbtYmSJLjGd9KTpk9J4H2jDZ",
                "5JcdQEQjBS92rKqwzQnpBndqieKAMQSiXLhU7SFZoCja5c1JyKM",
                "5JsmdqfNXegnM1eA8HyL6uimHp6pS9ba4kwoiWjjvqFC1fY5AeV",
                "5J2KeFptc73WTZPoT1Sd59prFep6SobGobCYm7T5ZnBKtuW9RL9",
                "5HryThsy6ySbkaiGK12r8kQ21vNdH81T5iifFEZNTe59wfPFvU9",
                "5Ji4N7LSSv3MAVkM3Gw2kq8GT5uxZYNaZ3d3y2C4Ex1m7vshjBN",
                "5HqSHfckRKmZLqqWW7p2iU18BYvyjxQs2sksRWhXMWXsNEtxPZU",
            ],
        )

    def test_BrainKey_normalize(self):
        b = "COLORER BICORN KASBEKE FAERIE LOCHIA GOMUTI SOVKHOZ Y GERMAL AUNTIE PERFUMY TIME FEATURE GANGAN CELEMIN MATZO"
        self.assertEqual(
            [
                BrainKey(b + "").get_brainkey(),
                BrainKey(b + " ").get_brainkey(),
                BrainKey(b + "  ").get_brainkey(),
                BrainKey(b + "\t").get_brainkey(),
                BrainKey(b + "\t\t").get_brainkey(),
                BrainKey(b.replace(" ", "\t")).get_brainkey(),
                BrainKey(b.replace(" ", "  ")).get_brainkey(),
            ],
            [b, b, b, b, b, b, b],
        )

    def test_BrainKey_suggest(self):
        b = BrainKey()
        self.assertTrue(len(b.suggest()) > 0)

    def test_child(self):
        p = PrivateKey("5JWcdkhL3w4RkVPcZMdJsjos22yB5cSkPExerktvKnRNZR5gx1S")
        p2 = p.child(b"Foobar")
        self.assertIsInstance(p2, PrivateKey)
        self.assertEqual(str(p2), "5JQ6AQmjpbEZjJBLnoa3BaWa9y3LDTUBeSDwEGQD2UjYkb1gY2x")

    def test_BrainKey_sequences(self):
        b = BrainKey(
            "COLORER BICORN KASBEKE FAERIE LOCHIA GOMUTI SOVKHOZ Y GERMAL AUNTIE PERFUMY TIME FEATURE GANGAN CELEMIN MATZO"
        )
        keys = [
            "5Hsbn6kXio4bb7eW5bX7kTp2sdkmbzP8kGWoau46Cf7en7T1RRE",
            "5K9MHEyiSye5iFL2srZu3ZVjzAZjcQxUgUvuttcVrymovFbU4cc",
            "5JBXhzDWQdYPAzRxxuGtzqM7ULLKPK7GZmktHTyF9foGGfbtDLT",
            "5Kbbfbs6DmJFNddWiP1XZfDKwhm5dkn9KX5AENQfQke2RYBBDcz",
            "5JUqLwgxn8f7myNz4gDwo5e77HZgopHMDHv4icNVww9Rxu1GDG5",
            "5JNBVj5QVh86N8MUUwY3EVUmsZwChZftxnuJx22DzEtHWC4rmvK",
            "5JdvczYtxPPjQdXMki1tpNvuSbvPMxJG5y4ndEAuQsC5RYMQXuC",
            "5HsUSesU2YB4EA3dmpGtHh8aPAwEdkdhidG8hcU2Nd2tETKk85t",
            "5JpveiQd1mt91APyQwvsCdAXWJ7uag3JmhtSxpGienic8vv1k2W",
            "5KDGhQUqQmwcGQ9tegimSyyT4vmH8h2fMzoNe1MT9bEGvRvR6kD",
        ]
        for i in keys:
            p = b.next_sequence().get_private()
            self.assertEqual(str(p), i)

    def test_PasswordKey(self):
        a = [
            "Aang7foN3oz1Ungai2qua5toh3map8ladei1eem2ohsh2shuo8aeji9Thoseo7ah",
            "iep1Mees9eghiifahwei5iidi0Sazae9aigaeT7itho3quoo2dah5zuvobaelau5",
            "ohBeuyoothae5aer9odaegh5Eeloh1fi7obei9ahSh0haeYuas1sheehaiv5LaiX",
            "geiQuoo9NeeLoaZee0ain3Ku1biedohsesien4uHo1eib1ahzaesh5shae3iena7",
            "jahzeice6Ix8ohBo3eik9pohjahgeegoh9sahthai1aeMahs8ki7Iub1oojeeSuo",
            "eiVahHoh2hi4fazah9Tha8loxeeNgequaquuYee6Shoopo3EiWoosheeX6yohg2o",
            "PheeCh3ar8xoofoiphoo4aisahjiiPah4vah0eeceiJ2iyeem9wahyupeithah9T",
            "IuyiibahNgieshei2eeFu8aic1IeMae9ooXi9jaiwaht4Wiengieghahnguang0U",
            "Ipee1quee7sheughemae4eir8pheix3quac3ei0Aquo9ohieLaeseeh8AhGeM2ew",
            "Tech5iir0aP6waiMeiHoph3iwoch4iijoogh0zoh9aSh6Ueb2Dee5dang1aa8IiP",
        ]
        b = [
            "STM5NyCrrXHmdikC6QPRAPoDjSHVQJe3WC5bMZuF6YhqhSsfYfjhN",
            "STM8gyvJtYyv5ZbT2ZxbAtgufQ5ovV2bq6EQp4YDTzQuSwyg7Ckry",
            "STM7yE71iVPSpaq8Ae2AmsKfyFxA8pwYv5zgQtCnX7xMwRUQMVoGf",
            "STM5jRgWA2kswPaXsQNtD2MMjs92XfJ1TYob6tjHtsECg2AusF5Wo",
            "STM6XHwVxcP6zP5NV1jUbG6Kso9m8ZG9g2CjDiPcZpAxHngx6ATPB",
            "STM59X1S4ofTAeHd1iNHDGxim5GkLo2AdcznksUsSYGU687ywB5WV",
            "STM6BPPL4iSRbFVVN8v3BEEEyDsC1STRK7Ba9ewQ4Lqvszn5J8VAe",
            "STM7cdK927wj95ptUrCk6HKWVeF74LG5cTjDTV22Z3yJ4Xw8xc9qp",
            "STM7VNFRjrE1hs1CKpEAP9NAabdFpwvzYXRKvkrVBBv2kTQCbNHz7",
            "STM7ZZFhEBjujcKjkmY31i1spPMx6xDSRhkursZLigi2HKLuALe5t",
        ]
        for i, pwd in enumerate(a):
            p = format(PasswordKey("xeroc", pwd, "posting").get_public(), "STM")
            self.assertEqual(p, b[i])

    def test_Privatekey_pubkey(self):
        self.assertEqual(
            [
                format(
                    PrivateKey("5HvVz6XMx84aC5KaaBbwYrRLvWE46cH6zVnv4827SBPLorg76oq").pubkey, "STX"
                ),
                str(
                    PrivateKey(
                        "5HvVz6XMx84aC5KaaBbwYrRLvWE46cH6zVnv4827SBPLorg76oq", prefix="STX"
                    ).pubkey
                ),
            ],
            [
                "STX7W5qsanXHgRAZPijbrLMDwX6VmHqUdL2s8PZiYKD5h1R7JaqRJ",
                "STX7W5qsanXHgRAZPijbrLMDwX6VmHqUdL2s8PZiYKD5h1R7JaqRJ",
            ],
        )

    def test_Privatekey_derive(self):
        p1 = PrivateKey("5HvVz6XMx84aC5KaaBbwYrRLvWE46cH6zVnv4827SBPLorg76oq")
        p2 = PrivateKey("5Jete5oFNjjk3aUMkKuxgAXsp7ZyhgJbYNiNjHLvq5xzXkiqw7R")
        self.assertEqual(
            [
                format(p1.child(p2.get_secret()), "STM"),
                format(p2.child(p1.get_secret()), "STM"),
                format(p1.derive_private_key(0), "STM"),
                format(p2.derive_private_key(56), "STM"),
            ],
            [
                "STMZiwJpC7MUmc9gn3vii3XS36nUceYEfKvFC1NLSrjB7ZRQJ7gt",
                "STM24hzNSDZYgm9C85yxJqyk32DwjXg8pCgkGVzB77hvP2XxGDdvr",
                "STM2e99iqVQUFij7Dk2nWVNC1dL8M86q37Nj4KwPHKBu1Yy49HkwA",
                "STMgqaH9RdvUtVk7NFnx4BZJRrNS7Lj35qaueAeYJ3tKEqPaLwa4",
            ],
        )

    def test_utf8_nfkd(self):
        # The same sentence in various UTF-8 forms
        words_nfkd = "Pr\u030ci\u0301s\u030cerne\u030c z\u030clut\u030couc\u030cky\u0301 ku\u030an\u030c u\u0301pe\u030cl d\u030ca\u0301belske\u0301 o\u0301dy za\u0301ker\u030cny\u0301 uc\u030cen\u030c be\u030cz\u030ci\u0301 pode\u0301l zo\u0301ny u\u0301lu\u030a"
        words_nfc = "P\u0159\xed\u0161ern\u011b \u017elu\u0165ou\u010dk\xfd k\u016f\u0148 \xfap\u011bl \u010f\xe1belsk\xe9 \xf3dy z\xe1ke\u0159n\xfd u\u010de\u0148 b\u011b\u017e\xed pod\xe9l z\xf3ny \xfal\u016f"
        words_nfkc = "P\u0159\xed\u0161ern\u011b \u017elu\u0165ou\u010dk\xfd k\u016f\u0148 \xfap\u011bl \u010f\xe1belsk\xe9 \xf3dy z\xe1ke\u0159n\xfd u\u010de\u0148 b\u011b\u017e\xed pod\xe9l z\xf3ny \xfal\u016f"
        words_nfd = "Pr\u030ci\u0301s\u030cerne\u030c z\u030clut\u030couc\u030cky\u0301 ku\u030an\u030c u\u0301pe\u030cl d\u030ca\u0301belske\u0301 o\u0301dy za\u0301ker\u030cny\u0301 uc\u030cen\u030c be\u030cz\u030ci\u0301 pode\u0301l zo\u0301ny u\u0301lu\u030a"

        passphrase_nfkd = "Neuve\u030cr\u030citelne\u030c bezpec\u030cne\u0301 hesli\u0301c\u030cko"
        passphrase_nfc = "Neuv\u011b\u0159iteln\u011b bezpe\u010dn\xe9 hesl\xed\u010dko"
        passphrase_nfkc = "Neuv\u011b\u0159iteln\u011b bezpe\u010dn\xe9 hesl\xed\u010dko"
        passphrase_nfd = "Neuve\u030cr\u030citelne\u030c bezpec\u030cne\u0301 hesli\u0301c\u030cko"

        seed_nfkd = Mnemonic.to_seed(words_nfkd, passphrase_nfkd)
        seed_nfc = Mnemonic.to_seed(words_nfc, passphrase_nfc)
        seed_nfkc = Mnemonic.to_seed(words_nfkc, passphrase_nfkc)
        seed_nfd = Mnemonic.to_seed(words_nfd, passphrase_nfd)

        self.assertEqual(seed_nfkd, seed_nfc)
        self.assertEqual(seed_nfkd, seed_nfkc)
        self.assertEqual(seed_nfkd, seed_nfd)

    def test_expand(self):
        m = Mnemonic()
        self.assertEqual("access", m.expand("access"))
        self.assertEqual(
            "access access acb acc act action", m.expand("access acce acb acc act acti")
        )

    def test_expand_word(self):
        m = Mnemonic()
        self.assertEqual("", m.expand_word(""))
        self.assertEqual(" ", m.expand_word(" "))
        self.assertEqual("access", m.expand_word("access"))  # word in list
        self.assertEqual("access", m.expand_word("acce"))  # unique prefix expanded to word in list
        self.assertEqual("acb", m.expand_word("acb"))  # not found at all
        self.assertEqual("acc", m.expand_word("acc"))  # multi-prefix match
        self.assertEqual("act", m.expand_word("act"))  # exact three letter match
        self.assertEqual("action", m.expand_word("acti"))  # unique prefix expanded to word in list

    def test_failed_checksum(self):
        code = "bless cloud wheel regular tiny venue bird web grief security dignity zoo"
        mnemo = Mnemonic()
        self.assertFalse(mnemo.check(code))

    def test_mnemonic(self):
        v = [
            [
                "00000000000000000000000000000000",
                "abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon about",
                "c55257c360c07c72029aebc1b53c05ed0362ada38ead3e3e9efa3708e53495531f09a6987599d18264c1e1c92f2cf141630c7a3c4ab7c81b2f001698e7463b04",
                "xprv9s21ZrQH143K3h3fDYiay8mocZ3afhfULfb5GX8kCBdno77K4HiA15Tg23wpbeF1pLfs1c5SPmYHrEpTuuRhxMwvKDwqdKiGJS9XFKzUsAF",
            ],
            [
                "7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f",
                "legal winner thank year wave sausage worth useful legal winner thank yellow",
                "2e8905819b8723fe2c1d161860e5ee1830318dbf49a83bd451cfb8440c28bd6fa457fe1296106559a3c80937a1c1069be3a3a5bd381ee6260e8d9739fce1f607",
                "xprv9s21ZrQH143K2gA81bYFHqU68xz1cX2APaSq5tt6MFSLeXnCKV1RVUJt9FWNTbrrryem4ZckN8k4Ls1H6nwdvDTvnV7zEXs2HgPezuVccsq",
            ],
            [
                "80808080808080808080808080808080",
                "letter advice cage absurd amount doctor acoustic avoid letter advice cage above",
                "d71de856f81a8acc65e6fc851a38d4d7ec216fd0796d0a6827a3ad6ed5511a30fa280f12eb2e47ed2ac03b5c462a0358d18d69fe4f985ec81778c1b370b652a8",
                "xprv9s21ZrQH143K2shfP28KM3nr5Ap1SXjz8gc2rAqqMEynmjt6o1qboCDpxckqXavCwdnYds6yBHZGKHv7ef2eTXy461PXUjBFQg6PrwY4Gzq",
            ],
            [
                "ffffffffffffffffffffffffffffffff",
                "zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo wrong",
                "ac27495480225222079d7be181583751e86f571027b0497b5b5d11218e0a8a13332572917f0f8e5a589620c6f15b11c61dee327651a14c34e18231052e48c069",
                "xprv9s21ZrQH143K2V4oox4M8Zmhi2Fjx5XK4Lf7GKRvPSgydU3mjZuKGCTg7UPiBUD7ydVPvSLtg9hjp7MQTYsW67rZHAXeccqYqrsx8LcXnyd",
            ],
            [
                "000000000000000000000000000000000000000000000000",
                "abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon agent",
                "035895f2f481b1b0f01fcf8c289c794660b289981a78f8106447707fdd9666ca06da5a9a565181599b79f53b844d8a71dd9f439c52a3d7b3e8a79c906ac845fa",
                "xprv9s21ZrQH143K3mEDrypcZ2usWqFgzKB6jBBx9B6GfC7fu26X6hPRzVjzkqkPvDqp6g5eypdk6cyhGnBngbjeHTe4LsuLG1cCmKJka5SMkmU",
            ],
            [
                "7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f",
                "legal winner thank year wave sausage worth useful legal winner thank year wave sausage worth useful legal will",
                "f2b94508732bcbacbcc020faefecfc89feafa6649a5491b8c952cede496c214a0c7b3c392d168748f2d4a612bada0753b52a1c7ac53c1e93abd5c6320b9e95dd",
                "xprv9s21ZrQH143K3Lv9MZLj16np5GzLe7tDKQfVusBni7toqJGcnKRtHSxUwbKUyUWiwpK55g1DUSsw76TF1T93VT4gz4wt5RM23pkaQLnvBh7",
            ],
            [
                "808080808080808080808080808080808080808080808080",
                "letter advice cage absurd amount doctor acoustic avoid letter advice cage absurd amount doctor acoustic avoid letter always",
                "107d7c02a5aa6f38c58083ff74f04c607c2d2c0ecc55501dadd72d025b751bc27fe913ffb796f841c49b1d33b610cf0e91d3aa239027f5e99fe4ce9e5088cd65",
                "xprv9s21ZrQH143K3VPCbxbUtpkh9pRG371UCLDz3BjceqP1jz7XZsQ5EnNkYAEkfeZp62cDNj13ZTEVG1TEro9sZ9grfRmcYWLBhCocViKEJae",
            ],
            [
                "ffffffffffffffffffffffffffffffffffffffffffffffff",
                "zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo when",
                "0cd6e5d827bb62eb8fc1e262254223817fd068a74b5b449cc2f667c3f1f985a76379b43348d952e2265b4cd129090758b3e3c2c49103b5051aac2eaeb890a528",
                "xprv9s21ZrQH143K36Ao5jHRVhFGDbLP6FCx8BEEmpru77ef3bmA928BxsqvVM27WnvvyfWywiFN8K6yToqMaGYfzS6Db1EHAXT5TuyCLBXUfdm",
            ],
            [
                "0000000000000000000000000000000000000000000000000000000000000000",
                "abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon art",
                "bda85446c68413707090a52022edd26a1c9462295029f2e60cd7c4f2bbd3097170af7a4d73245cafa9c3cca8d561a7c3de6f5d4a10be8ed2a5e608d68f92fcc8",
                "xprv9s21ZrQH143K32qBagUJAMU2LsHg3ka7jqMcV98Y7gVeVyNStwYS3U7yVVoDZ4btbRNf4h6ibWpY22iRmXq35qgLs79f312g2kj5539ebPM",
            ],
            [
                "7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f",
                "legal winner thank year wave sausage worth useful legal winner thank year wave sausage worth useful legal winner thank year wave sausage worth title",
                "bc09fca1804f7e69da93c2f2028eb238c227f2e9dda30cd63699232578480a4021b146ad717fbb7e451ce9eb835f43620bf5c514db0f8add49f5d121449d3e87",
                "xprv9s21ZrQH143K3Y1sd2XVu9wtqxJRvybCfAetjUrMMco6r3v9qZTBeXiBZkS8JxWbcGJZyio8TrZtm6pkbzG8SYt1sxwNLh3Wx7to5pgiVFU",
            ],
            [
                "8080808080808080808080808080808080808080808080808080808080808080",
                "letter advice cage absurd amount doctor acoustic avoid letter advice cage absurd amount doctor acoustic avoid letter advice cage absurd amount doctor acoustic bless",
                "c0c519bd0e91a2ed54357d9d1ebef6f5af218a153624cf4f2da911a0ed8f7a09e2ef61af0aca007096df430022f7a2b6fb91661a9589097069720d015e4e982f",
                "xprv9s21ZrQH143K3CSnQNYC3MqAAqHwxeTLhDbhF43A4ss4ciWNmCY9zQGvAKUSqVUf2vPHBTSE1rB2pg4avopqSiLVzXEU8KziNnVPauTqLRo",
            ],
            [
                "ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff",
                "zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo vote",
                "dd48c104698c30cfe2b6142103248622fb7bb0ff692eebb00089b32d22484e1613912f0a5b694407be899ffd31ed3992c456cdf60f5d4564b8ba3f05a69890ad",
                "xprv9s21ZrQH143K2WFF16X85T2QCpndrGwx6GueB72Zf3AHwHJaknRXNF37ZmDrtHrrLSHvbuRejXcnYxoZKvRquTPyp2JiNG3XcjQyzSEgqCB",
            ],
            [
                "9e885d952ad362caeb4efe34a8e91bd2",
                "ozone drill grab fiber curtain grace pudding thank cruise elder eight picnic",
                "274ddc525802f7c828d8ef7ddbcdc5304e87ac3535913611fbbfa986d0c9e5476c91689f9c8a54fd55bd38606aa6a8595ad213d4c9c9f9aca3fb217069a41028",
                "xprv9s21ZrQH143K2oZ9stBYpoaZ2ktHj7jLz7iMqpgg1En8kKFTXJHsjxry1JbKH19YrDTicVwKPehFKTbmaxgVEc5TpHdS1aYhB2s9aFJBeJH",
            ],
            [
                "6610b25967cdcca9d59875f5cb50b0ea75433311869e930b",
                "gravity machine north sort system female filter attitude volume fold club stay feature office ecology stable narrow fog",
                "628c3827a8823298ee685db84f55caa34b5cc195a778e52d45f59bcf75aba68e4d7590e101dc414bc1bbd5737666fbbef35d1f1903953b66624f910feef245ac",
                "xprv9s21ZrQH143K3uT8eQowUjsxrmsA9YUuQQK1RLqFufzybxD6DH6gPY7NjJ5G3EPHjsWDrs9iivSbmvjc9DQJbJGatfa9pv4MZ3wjr8qWPAK",
            ],
            [
                "68a79eaca2324873eacc50cb9c6eca8cc68ea5d936f98787c60c7ebc74e6ce7c",
                "hamster diagram private dutch cause delay private meat slide toddler razor book happy fancy gospel tennis maple dilemma loan word shrug inflict delay length",
                "64c87cde7e12ecf6704ab95bb1408bef047c22db4cc7491c4271d170a1b213d20b385bc1588d9c7b38f1b39d415665b8a9030c9ec653d75e65f847d8fc1fc440",
                "xprv9s21ZrQH143K2XTAhys3pMNcGn261Fi5Ta2Pw8PwaVPhg3D8DWkzWQwjTJfskj8ofb81i9NP2cUNKxwjueJHHMQAnxtivTA75uUFqPFeWzk",
            ],
            [
                "c0ba5a8e914111210f2bd131f3d5e08d",
                "scheme spot photo card baby mountain device kick cradle pact join borrow",
                "ea725895aaae8d4c1cf682c1bfd2d358d52ed9f0f0591131b559e2724bb234fca05aa9c02c57407e04ee9dc3b454aa63fbff483a8b11de949624b9f1831a9612",
                "xprv9s21ZrQH143K3FperxDp8vFsFycKCRcJGAFmcV7umQmcnMZaLtZRt13QJDsoS5F6oYT6BB4sS6zmTmyQAEkJKxJ7yByDNtRe5asP2jFGhT6",
            ],
            [
                "6d9be1ee6ebd27a258115aad99b7317b9c8d28b6d76431c3",
                "horn tenant knee talent sponsor spell gate clip pulse soap slush warm silver nephew swap uncle crack brave",
                "fd579828af3da1d32544ce4db5c73d53fc8acc4ddb1e3b251a31179cdb71e853c56d2fcb11aed39898ce6c34b10b5382772db8796e52837b54468aeb312cfc3d",
                "xprv9s21ZrQH143K3R1SfVZZLtVbXEB9ryVxmVtVMsMwmEyEvgXN6Q84LKkLRmf4ST6QrLeBm3jQsb9gx1uo23TS7vo3vAkZGZz71uuLCcywUkt",
            ],
            [
                "9f6a2878b2520799a44ef18bc7df394e7061a224d2c33cd015b157d746869863",
                "panda eyebrow bullet gorilla call smoke muffin taste mesh discover soft ostrich alcohol speed nation flash devote level hobby quick inner drive ghost inside",
                "72be8e052fc4919d2adf28d5306b5474b0069df35b02303de8c1729c9538dbb6fc2d731d5f832193cd9fb6aeecbc469594a70e3dd50811b5067f3b88b28c3e8d",
                "xprv9s21ZrQH143K2WNnKmssvZYM96VAr47iHUQUTUyUXH3sAGNjhJANddnhw3i3y3pBbRAVk5M5qUGFr4rHbEWwXgX4qrvrceifCYQJbbFDems",
            ],
            [
                "23db8160a31d3e0dca3688ed941adbf3",
                "cat swing flag economy stadium alone churn speed unique patch report train",
                "deb5f45449e615feff5640f2e49f933ff51895de3b4381832b3139941c57b59205a42480c52175b6efcffaa58a2503887c1e8b363a707256bdd2b587b46541f5",
                "xprv9s21ZrQH143K4G28omGMogEoYgDQuigBo8AFHAGDaJdqQ99QKMQ5J6fYTMfANTJy6xBmhvsNZ1CJzRZ64PWbnTFUn6CDV2FxoMDLXdk95DQ",
            ],
            [
                "8197a4a47f0425faeaa69deebc05ca29c0a5b5cc76ceacc0",
                "light rule cinnamon wrap drastic word pride squirrel upgrade then income fatal apart sustain crack supply proud access",
                "4cbdff1ca2db800fd61cae72a57475fdc6bab03e441fd63f96dabd1f183ef5b782925f00105f318309a7e9c3ea6967c7801e46c8a58082674c860a37b93eda02",
                "xprv9s21ZrQH143K3wtsvY8L2aZyxkiWULZH4vyQE5XkHTXkmx8gHo6RUEfH3Jyr6NwkJhvano7Xb2o6UqFKWHVo5scE31SGDCAUsgVhiUuUDyh",
            ],
            [
                "066dca1a2bb7e8a1db2832148ce9933eea0f3ac9548d793112d9a95c9407efad",
                "all hour make first leader extend hole alien behind guard gospel lava path output census museum junior mass reopen famous sing advance salt reform",
                "26e975ec644423f4a4c4f4215ef09b4bd7ef924e85d1d17c4cf3f136c2863cf6df0a475045652c57eb5fb41513ca2a2d67722b77e954b4b3fc11f7590449191d",
                "xprv9s21ZrQH143K3rEfqSM4QZRVmiMuSWY9wugscmaCjYja3SbUD3KPEB1a7QXJoajyR2T1SiXU7rFVRXMV9XdYVSZe7JoUXdP4SRHTxsT1nzm",
            ],
            [
                "f30f8c1da665478f49b001d94c5fc452",
                "vessel ladder alter error federal sibling chat ability sun glass valve picture",
                "2aaa9242daafcee6aa9d7269f17d4efe271e1b9a529178d7dc139cd18747090bf9d60295d0ce74309a78852a9caadf0af48aae1c6253839624076224374bc63f",
                "xprv9s21ZrQH143K2QWV9Wn8Vvs6jbqfF1YbTCdURQW9dLFKDovpKaKrqS3SEWsXCu6ZNky9PSAENg6c9AQYHcg4PjopRGGKmdD313ZHszymnps",
            ],
            [
                "c10ec20dc3cd9f652c7fac2f1230f7a3c828389a14392f05",
                "scissors invite lock maple supreme raw rapid void congress muscle digital elegant little brisk hair mango congress clump",
                "7b4a10be9d98e6cba265566db7f136718e1398c71cb581e1b2f464cac1ceedf4f3e274dc270003c670ad8d02c4558b2f8e39edea2775c9e232c7cb798b069e88",
                "xprv9s21ZrQH143K4aERa2bq7559eMCCEs2QmmqVjUuzfy5eAeDX4mqZffkYwpzGQRE2YEEeLVRoH4CSHxianrFaVnMN2RYaPUZJhJx8S5j6puX",
            ],
            [
                "f585c11aec520db57dd353c69554b21a89b20fb0650966fa0a9d6f74fd989d8f",
                "void come effort suffer camp survey warrior heavy shoot primary clutch crush open amazing screen patrol group space point ten exist slush involve unfold",
                "01f5bced59dec48e362f2c45b5de68b9fd6c92c6634f44d6d40aab69056506f0e35524a518034ddc1192e1dacd32c1ed3eaa3c3b131c88ed8e7e54c49a5d0998",
                "xprv9s21ZrQH143K39rnQJknpH1WEPFJrzmAqqasiDcVrNuk926oizzJDDQkdiTvNPr2FYDYzWgiMiC63YmfPAa2oPyNB23r2g7d1yiK6WpqaQS",
            ],
        ]
        mnemo = Mnemonic()
        for i in range(len(v)):
            code = mnemo.to_mnemonic(unhexlify(v[i][0]))
            seed = Mnemonic.to_seed(code, passphrase="TREZOR")
            key = BIP32Key.fromEntropy(seed)
            if sys.version >= "3":
                seed = hexlify(seed).decode("utf8")
            self.assertIs(mnemo.check(v[i][1]), True)
            self.assertEqual(v[i][1], code)
            self.assertEqual(v[i][2], seed)
            xprv = key.ExtendedKey(private=True, encoded=True)
            self.assertEqual(v[i][3], xprv)

    def test_to_entropy(self):
        data = [bytearray((random.getrandbits(8) for _ in range(32))) for _ in range(1024)]
        data.append(b"Lorem ipsum dolor sit amet amet.")
        m = Mnemonic()
        for d in data:
            self.assertEqual(m.to_entropy(m.to_mnemonic(d).split()), d)

    def test_mnemorickey(self):
        word_list = "news clever spot drama infant detail sword cover color throw foot primary when slender rhythm clog autumn ecology enough bronze math you modify excuse"
        mk = MnemonicKey(word_list)
        mk.set_path_BIP48(role="owner")
        self.assertEqual(
            str(mk.get_private_key()),
            str(PrivateKey("L2cgypn75kre1s7JUkTK6H7Y656GwDbNnSNZKWSQ2Rnnx6qM11KD")),
        )
        self.assertEqual(
            str(mk.get_public_key()),
            str(PublicKey("02821aa2d26c4fd9b735dd1fed148b96fec978eae1440adf79a4bf95e118b2d8f1")),
        )

        mk = MnemonicKey(word_list)
        mk.set_path_BIP48(role="owner", key_sequence=5)
        self.assertEqual(
            str(mk.get_private_key()),
            str(PrivateKey("L5cJSZPcBMBtmuRaK9C48yyXK5JpaH15BsjKLZkmamEWKKx25Kx7")),
        )
        self.assertEqual(
            str(mk.get_public_key()),
            str(PublicKey("0309a45aa9add2c7421e0553e23b1800e51d95e525fa4eae1bcc5fb58186e07ed5")),
        )

        mk = MnemonicKey(word_list)
        mk.set_path_BIP48(role="owner", account_sequence=2)
        self.assertEqual(
            str(mk.get_private_key()),
            str(PrivateKey("L4A95nfp1kyJtUtaTqzMyLQkz6NYSfk4R8pejcgKXQSUtPysiFgv")),
        )
        self.assertEqual(
            str(mk.get_public_key()),
            str(PublicKey("02b164dc8819830cd50fca4217ad35fa7371cf29db1bc6a07456cc0090ca8ea8fe")),
        )

        mk = MnemonicKey(word_list)
        mk.set_path_BIP48(role="active")
        self.assertEqual(
            str(mk.get_private_key()),
            str(PrivateKey("KygWePihetfhKYCHahcHjMebNSy53HtcXkccYuTn6R8QLydyPUWt")),
        )
        self.assertEqual(
            str(mk.get_public_key()),
            str(PublicKey("035c679454155c4c41e8956ecb8e514d37d2d28da91db81c3a22f216a09af94605")),
        )

        mk = MnemonicKey(word_list)
        mk.set_path_BIP48(role="posting")
        self.assertEqual(
            str(mk.get_private_key()),
            str(PrivateKey("L53K986B756VbqatsCi3jjPLHCq8Y38AZyXf19w6CcxuGH84Rrhs")),
        )
        self.assertEqual(
            str(mk.get_public_key()),
            str(PublicKey("0200e7d987dfd5aaecf822420475ddcbadc8503a99893d50d06f87da48e85a8206")),
        )

        mk = MnemonicKey(word_list)
        mk.set_path_BIP48(role="memo")
        self.assertEqual(
            str(mk.get_private_key()),
            str(PrivateKey("L5GrFqdRsroM1Ym4aMdALQBL7xN9kNMru9JTgtwbHVZ4iGvx1184")),
        )
        self.assertEqual(
            str(mk.get_public_key()),
            str(PublicKey("02fa2cdf5a007b01b1911615a4fba9c2a864a1c1ed079d222e5d549d207412c601")),
        )

    def test_new_privatekey(self):
        w = PrivateKey()
        self.assertIsInstance(w, PrivateKey)
        self.assertEqual(str(w)[0], "5")  # is a wif key that starts with 5

    def test_new_BrainKey(self):
        w = BrainKey().get_private_key()
        self.assertIsInstance(w, PrivateKey)
        self.assertEqual(str(w)[0], "5")  # is a wif key that starts with 5

    def test_derive_private_key(self):
        p = PrivateKey("5JWcdkhL3w4RkVPcZMdJsjos22yB5cSkPExerktvKnRNZR5gx1S")
        p2 = p.derive_private_key(10)
        self.assertEqual(
            repr(p2), "2dc7cb99933132e25b37710f9ea806228b04a583da11a137ef97fd42c0007390"
        )

    def test_derive_child(self):
        # NOTE: this key + offset pair is particularly nasty, as
        # the resulting derived value is less then 64 bytes long.
        # Thus, this test also tests for proper padding.
        p = PrivateKey("5K6hMUtQB2xwjuz3SRR6uM5HNERWgBqcK7gPPZ31XtAyBNoATZd")
        p2 = p.child(
            b"\xaf\x8f: \xf6T?V\x0bM\xd8\x16 \xfd\xde\xe9\xb9\xac\x03\r\xba\xb2\x8d\x868-\xc2\x90\x80\xe8\x1b\xce"
        )
        self.assertEqual(
            repr(p2), "0c5fae344a513a4cfab312b24c08df2b2d6afa25c0ead0d3d1d0d3e76794109b"
        )



# -*- coding: utf-8 -*-
import unittest
from decimal import Decimal

from nectar import Hive
from nectar.amount import Amount
from nectar.asset import Asset
from nectar.instance import set_shared_blockchain_instance

from .nodes import get_hive_nodes


class Testcases(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        """
        Initialize class-level test fixtures.

        Creates a Hive blockchain instance configured for testing (no network broadcast, with retries), registers it as the shared blockchain instance, and prepares Asset objects used across tests. Sets the following class attributes for use by tests:
        - bts: Hive instance
        - asset: Asset("HBD")
        - symbol: symbol string for the HBD asset
        - precision: precision (integer) for the HBD asset
        - asset2: Asset("HIVE")

        This method has the side effect of modifying global/shared blockchain state via set_shared_blockchain_instance.
        """
        cls.bts = Hive(node=get_hive_nodes(), nobroadcast=True, num_retries=10)
        set_shared_blockchain_instance(cls.bts)
        cls.asset = Asset("HBD")
        cls.symbol = cls.asset["symbol"]
        cls.precision = cls.asset["precision"]
        cls.asset2 = Asset("HIVE")

    def dotest(self, ret, amount, symbol):
        """
        Helper assertion for tests that validates an Amount-like result.

        Asserts that `ret` represents the expected numeric value and asset:
        - numeric value matches `amount` when converted to float,
        - `ret["symbol"]` equals `symbol`,
        - `ret["asset"]` is a dict,
        - `ret["amount"]` is a Decimal.

        Parameters:
            ret: Mapping-like object returned by Amount.json() or similar representation.
            amount: Expected numeric amount (numeric or Decimal-compatible).
            symbol: Expected asset symbol string.

        Raises:
            AssertionError: If any of the assertions fail.
        """
        self.assertEqual(float(ret), float(amount))
        self.assertEqual(ret["symbol"], symbol)
        self.assertIsInstance(ret["asset"], dict)
        self.assertIsInstance(ret["amount"], Decimal)

    def test_init(self):
        hv = self.bts
        # String init
        asset = Asset("HBD", blockchain_instance=hv)
        symbol = asset["symbol"]
        precision = asset["precision"]
        amount = Amount("1 {}".format(symbol), blockchain_instance=hv)
        self.dotest(amount, 1, symbol)

        # Amount init
        amount = Amount(amount, blockchain_instance=hv)
        self.dotest(amount, 1, symbol)

        # blockchain dict init
        amount = Amount(
            {"amount": 1 * 10**precision, "asset_id": asset["id"]}, blockchain_instance=hv
        )
        self.dotest(amount, 1, symbol)

        # API dict init
        amount = Amount(
            {"amount": 1.3 * 10**precision, "asset": asset["id"]}, blockchain_instance=hv
        )
        self.dotest(amount, 1.3, symbol)

        # Asset as symbol
        amount = Amount(1.3, Asset("HBD"), blockchain_instance=hv)
        self.dotest(amount, 1.3, symbol)

        # Asset as symbol
        amount = Amount(1.3, symbol, blockchain_instance=hv)
        self.dotest(amount, 1.3, symbol)

        # keyword inits
        amount = Amount(
            amount=1.3, asset=Asset("HBD", blockchain_instance=hv), blockchain_instance=hv
        )
        self.dotest(amount, 1.3, symbol)

        amount = Amount(
            amount=1.3001, asset=Asset("HBD", blockchain_instance=hv), blockchain_instance=hv
        )
        self.dotest(amount, 1.3001, symbol)

        amount = Amount(
            amount=1.3001,
            asset=Asset("HBD", blockchain_instance=hv),
            fixed_point_arithmetic=True,
            blockchain_instance=hv,
        )
        self.dotest(amount, 1.3, symbol)

        # keyword inits
        amount = Amount(
            amount=1.3, asset=dict(Asset("HBD", blockchain_instance=hv)), blockchain_instance=hv
        )
        self.dotest(amount, 1.3, symbol)

        # keyword inits
        amount = Amount(amount=1.3, asset=symbol, blockchain_instance=hv)
        self.dotest(amount, 1.3, symbol)

        amount = Amount(amount=8.190, asset=symbol, blockchain_instance=hv)
        self.dotest(amount, 8.190, symbol)

    def test_copy(self):
        amount = Amount("1", self.symbol)
        self.dotest(amount.copy(), 1, self.symbol)

    def test_properties(self):
        amount = Amount("1", self.symbol)
        self.assertEqual(amount.amount, 1.0)
        self.assertEqual(amount.symbol, self.symbol)
        self.assertIsInstance(amount.asset, Asset)
        self.assertEqual(amount.asset["symbol"], self.symbol)

    def test_tuple(self):
        amount = Amount("1", self.symbol)
        self.assertEqual(amount.tuple(), (1.0, self.symbol))

    def test_json_appbase(self):
        asset = Asset("HBD", blockchain_instance=self.bts)
        amount = Amount("1", asset, new_appbase_format=False, blockchain_instance=self.bts)
        if self.bts.rpc.get_use_appbase():
            self.assertEqual(
                amount.json(), [str(1 * 10**asset.precision), asset.precision, asset.asset]
            )
        else:
            self.assertEqual(amount.json(), "1.000 HBD")

    def test_json_appbase2(self):
        asset = Asset("HBD", blockchain_instance=self.bts)
        amount = Amount("1", asset, new_appbase_format=True, blockchain_instance=self.bts)
        if self.bts.rpc.get_use_appbase():
            self.assertEqual(
                amount.json(),
                {
                    "amount": str(1 * 10**asset.precision),
                    "nai": asset.asset,
                    "precision": asset.precision,
                },
            )
        else:
            self.assertEqual(amount.json(), "1.000 HBD")

    def test_string(self):
        self.assertEqual(str(Amount("10000", self.symbol)), "10000.000 {}".format(self.symbol))

    def test_int(self):
        self.assertEqual(int(Amount("0.9999", self.symbol)), 999)
        self.assertEqual(int(Amount(0.151, self.symbol)), 151)
        self.assertEqual(int(Amount(8.190, self.symbol)), 8190)
        self.assertEqual(int(Amount(round(0.1509, 3), self.symbol)), 151)
        self.assertEqual(int(Amount(round(0.1509, 3), self.asset)), 151)
        self.assertEqual(int(Amount(int(1), self.symbol)), 1000)
        self.assertEqual(int(Amount(amount=round(0.1509, 3), asset=Asset("HBD"))), 151)

    def test_dict(self):
        self.assertEqual(int(Amount({"amount": "150", "nai": "@@000000021", "precision": 3})), 150)

    def test_float(self):
        self.assertEqual(float(Amount("1", self.symbol)), 1.00000)
        self.assertEqual(float(Amount(0.151, self.symbol)), 0.151)
        self.assertEqual(float(Amount(round(0.1509, 3), self.symbol)), 0.151)
        self.assertEqual(float(Amount(8.190, self.symbol)), 8.190)

    def test_plus(self):
        a1 = Amount(1, self.symbol)
        a2 = Amount(2, self.symbol)
        self.dotest(a1 + a2, 3, self.symbol)
        self.dotest(a1 + 2, 3, self.symbol)
        with self.assertRaises(Exception):
            a1 + Amount(1, asset=self.asset2)
        # inline
        a2 = Amount(2, self.symbol)
        a2 += a1
        self.dotest(a2, 3, self.symbol)
        a2 += 5
        self.dotest(a2, 8, self.symbol)
        a2 += Decimal(2)
        self.dotest(a2, 10, self.symbol)
        with self.assertRaises(Exception):
            a1 += Amount(1, asset=self.asset2)

    def test_minus(self):
        a1 = Amount(1, self.symbol)
        a2 = Amount(2, self.symbol)
        self.dotest(a1 - a2, -1, self.symbol)
        self.dotest(a1 - 5, -4, self.symbol)
        with self.assertRaises(Exception):
            a1 - Amount(1, asset=self.asset2)
        # inline
        a2 = Amount(2, self.symbol)
        a2 -= a1
        self.dotest(a2, 1, self.symbol)
        a2 -= 1
        self.dotest(a2, 0, self.symbol)
        self.dotest(a2 - 2, -2, self.symbol)
        with self.assertRaises(Exception):
            a1 -= Amount(1, asset=self.asset2)

    def test_mul(self):
        a1 = Amount(5, self.symbol)
        a2 = Amount(2, self.symbol)
        self.dotest(a1 * a2, 10, self.symbol)
        self.dotest(a1 * 3, 15, self.symbol)
        with self.assertRaises(Exception):
            a1 * Amount(1, asset=self.asset2)
        # inline
        a2 = Amount(2, self.symbol)
        a2 *= 5
        self.dotest(a2, 10, self.symbol)
        a2 = Amount(2, self.symbol)
        a2 *= a1
        self.dotest(a2, 10, self.symbol)
        with self.assertRaises(Exception):
            a1 *= Amount(2, asset=self.asset2)

    def test_div(self):
        a1 = Amount(15, self.symbol)
        self.dotest(a1 / 3, 5, self.symbol)
        self.dotest(a1 // 2, 7, self.symbol)
        with self.assertRaises(Exception):
            a1 / Amount(1, asset=self.asset2)
        # inline
        a2 = a1.copy()
        a2 /= 3
        self.dotest(a2, 5, self.symbol)
        a2 = a1.copy()
        a2 //= 2
        self.dotest(a2, 7, self.symbol)
        with self.assertRaises(Exception):
            a1 *= Amount(2, asset=self.asset2)

    def test_mod(self):
        a1 = Amount(15, self.symbol)
        a2 = Amount(3, self.symbol)
        self.dotest(a1 % 3, 0, self.symbol)
        self.dotest(a1 % a2, 0, self.symbol)
        self.dotest(a1 % 2, 1, self.symbol)
        with self.assertRaises(Exception):
            a1 % Amount(1, asset=self.asset2)
        # inline
        a2 = a1.copy()
        a2 %= 3
        self.dotest(a2, 0, self.symbol)
        with self.assertRaises(Exception):
            a1 %= Amount(2, asset=self.asset2)

    def test_pow(self):
        a1 = Amount(15, self.symbol)
        a2 = Amount(3, self.symbol)
        self.dotest(a1**3, 15**3, self.symbol)
        self.dotest(a1**a2, 15**3, self.symbol)
        self.dotest(a1**2, 15**2, self.symbol)
        with self.assertRaises(Exception):
            a1 ** Amount(1, asset=self.asset2)
        # inline
        a2 = a1.copy()
        a2 **= 3
        self.dotest(a2, 15**3, self.symbol)
        with self.assertRaises(Exception):
            a1 **= Amount(2, asset=self.asset2)

    def test_ltge(self):
        a1 = Amount(1, self.symbol)
        a2 = Amount(2, self.symbol)
        self.assertTrue(a1 < a2)
        self.assertTrue(a2 > a1)
        self.assertTrue(a2 > 1)
        self.assertTrue(a1 < 5)

    def test_leeq(self):
        a1 = Amount(1, self.symbol)
        a2 = Amount(1, self.symbol)
        self.assertTrue(a1 <= a2)
        self.assertTrue(a1 >= a2)
        self.assertTrue(a1 <= 1)
        self.assertTrue(a1 >= 1)
        self.assertTrue(a1 == 1.0001)

    def test_ne(self):
        a1 = Amount(1, self.symbol)
        a2 = Amount(2, self.symbol)
        self.assertTrue(a1 != a2)
        self.assertTrue(a1 != 5)
        a1 = Amount(1, self.symbol)
        a2 = Amount(1, self.symbol)
        self.assertTrue(a1 == a2)
        self.assertTrue(a1 == 1)



# -*- coding: utf-8 -*-
import unittest

from nectar import Hive
from nectar.account import Account
from nectar.hivesigner import HiveSigner

from .nodes import get_hive_nodes

# Py3 compatibility
core_unit = "STM"


class Testcases(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        """
        Set up class-wide test fixtures: create a Hive blockchain instance and a test Account.

        Initializes a Hive client configured for local testing (no broadcast, unsigned transactions, long refresh interval, and retry settings) using get_hive_nodes(), and attaches an Account named "test" (full access) to that blockchain instance as class attributes.
        """
        cls.bts = Hive(
            node=get_hive_nodes(),
            nobroadcast=True,
            unsigned=True,
            data_refresh_time_seconds=900,
            num_retries=10,
        )

        cls.account = Account("test", full=True, blockchain_instance=cls.bts)

    def test_transfer(self):
        bts = self.bts
        acc = self.account
        acc.blockchain.txbuffer.clear()
        tx = acc.transfer("test1", 1.000, "HIVE", memo="test")
        sc2 = HiveSigner(blockchain_instance=bts)
        url = sc2.url_from_tx(tx)
        url_test = "https://hivesigner.com/oauth2/sign/transfer?from=test&to=test1&amount=1.000+HIVE&memo=test"
        print(f"Generated URL: {url} (length {len(url)})")
        print(f"Expected URL: {url_test} (length {len(url_test)})")
        self.assertEqual(len(url), len(url_test))
        self.assertEqual(len(url.split("?")), 2)
        self.assertEqual(url.split("?")[0], url_test.split("?")[0])
        # Compare query components irrespective of order
        url_parts = (url.split("?")[1]).split("&")
        url_test_parts = (url_test.split("?")[1]).split("&")
        self.assertEqual(len(url_parts), 4)
        self.assertEqual(len(list(set(url_parts).intersection(set(url_test_parts)))), 4)

    def test_login_url(self):
        bts = self.bts
        sc2 = HiveSigner(blockchain_instance=bts)
        url = sc2.get_login_url("localhost", scope="login,vote")
        url_test = "https://hivesigner.com/oauth2/authorize?client_id=None&redirect_uri=localhost&scope=login,vote"
        self.assertEqual(len(url), len(url_test))
        self.assertEqual(len(url.split("?")), 2)
        self.assertEqual(url.split("?")[0], url_test.split("?")[0])

    def test_sign_method(self):
        """Test the sign method functionality"""
        bts = self.bts
        sc2 = HiveSigner(blockchain_instance=bts)

        # Test transaction to sign
        test_tx = {
            "operations": [
                [
                    "vote",
                    {
                        "voter": "test",
                        "author": "gtg",
                        "permlink": "hive-pressure-4-need-for-speed",
                        "weight": 10000,
                    },
                ]
            ]
        }

        # Test signing
        signed_tx = sc2.sign(test_tx)

        # Verify the signed transaction structure
        self.assertIsInstance(signed_tx, dict)
        self.assertIn("operations", signed_tx)
        self.assertIn("signatures", signed_tx)
        self.assertEqual(len(signed_tx["signatures"]), 1)
        self.assertEqual(signed_tx["operations"], test_tx["operations"])

        # Test error cases
        with self.assertRaises(ValueError):
            sc2.sign("not a dict")

        with self.assertRaises(ValueError):
            sc2.sign({})

        with self.assertRaises(ValueError):
            sc2.sign({"operations": []})



# -*- coding: utf-8 -*-
import json

from nectargraphenebase.types import Id, JsonObj, Optional, String

from .operationids import operations


class Operation(object):
    def __init__(self, op):
        if isinstance(op, list) and len(op) == 2:
            if isinstance(op[0], int):
                self.opId = op[0]
                name = self.getOperationNameForId(self.opId)
            else:
                self.opId = self.operations().get(op[0], None)
                name = op[0]
                if self.opId is None:
                    raise ValueError("Unknown operation")
            self.name = name[0].upper() + name[1:]  # klassname
            try:
                klass = self._getklass(self.name)
            except Exception:
                raise NotImplementedError("Unimplemented Operation %s" % self.name)
            self.op = klass(op[1])
            self.appbase = False
        elif isinstance(op, dict):
            if len(op["type"]) > 10 and op["type"][-9:] == "operation":
                name = op["type"][:-10]
            else:
                name = op["type"]
            self.opId = self.operations().get(name, None)
            if self.opId is None:
                raise ValueError("Unknown operation")
            self.name = name[0].upper() + name[1:]  # klassname
            try:
                klass = self._getklass(self.name)
            except Exception:
                raise NotImplementedError("Unimplemented Operation %s" % self.name)
            self.op = klass(op["value"])
            self.appbase = True
        else:
            self.op = op
            self.name = type(self.op).__name__.lower()  # also store name
            self.opId = self.operations()[self.name]

    def operations(self):
        return operations

    def getOperationNameForId(self, i):
        """Convert an operation id into the corresponding string"""
        for key in self.operations():
            if int(self.operations()[key]) is int(i):
                return key
        return "Unknown Operation ID %d" % i

    def _getklass(self, name):
        module = __import__("graphenebase.operations", fromlist=["operations"])
        class_ = getattr(module, name)
        return class_

    def __bytes__(self):
        return bytes(Id(self.opId)) + bytes(self.op)

    def __str__(self):
        return json.dumps([self.opId, self.op.toJson()])


class GrapheneObject(object):
    """Core abstraction class

    This class is used for any JSON reflected object in Graphene.

    * ``instance.__json__()``: encodes data into json format
    * ``bytes(instance)``: encodes data into wire format
    * ``str(instances)``: dumps json object as string

    """

    def __init__(self, data=None):
        self.data = data

    def __bytes__(self):
        if self.data is None:
            return bytes()
        b = b""
        for name, value in list(self.data.items()):
            if isinstance(value, str):
                b += bytes(value, "utf-8")
            else:
                b += bytes(value)
        return b

    def __json__(self):
        if self.data is None:
            return {}
        d = {}  # JSON output is *not* ordered
        for name, value in list(self.data.items()):
            if isinstance(value, Optional) and value.isempty():
                continue

            if isinstance(value, String):
                d.update({name: str(value)})
            else:
                try:
                    d.update({name: JsonObj(value)})
                except Exception:
                    d.update({name: value.__str__()})
        return d

    def __str__(self):
        return json.dumps(self.__json__())

    def toJson(self):
        return self.__json__()

    def json(self):
        return self.__json__()


def isArgsThisClass(self, args):
    return len(args) == 1 and type(args[0]).__name__ == type(self).__name__
